# -*- coding: utf-8 -*-

from collective.behavior.internalnumber.browser.settings import set_settings
from collective.iconifiedcategory.utils import calculate_category_id
from datetime import datetime
from datetime import timedelta
from imio.helpers.content import object_values
from plonemeeting.restapi.config import CONFIG_ID_ERROR
from plonemeeting.restapi.config import CONFIG_ID_NOT_FOUND_ERROR
from plonemeeting.restapi.services.add import ANNEX_CONTENT_CATEGORY_ERROR
from plonemeeting.restapi.services.add import ANNEX_DECISION_RELATED_NOT_ITEM_ERROR
from plonemeeting.restapi.services.add import IGNORE_VALIDATION_FOR_REQUIRED_ERROR
from plonemeeting.restapi.services.add import IGNORE_VALIDATION_FOR_VALUED_ERROR
from plonemeeting.restapi.services.add import IGNORE_VALIDATION_FOR_WARNING
from plonemeeting.restapi.services.add import OPTIONAL_FIELD_ERROR
from plonemeeting.restapi.services.add import OPTIONAL_FIELDS_WARNING
from plonemeeting.restapi.services.add import ORG_FIELD_VALUE_ERROR
from plonemeeting.restapi.testing import PM_REST_TEST_ADD_ANNEXES_PROFILE_FUNCTIONAL
from plonemeeting.restapi.testing import PM_REST_TEST_ADD_PROFILE_FUNCTIONAL
from plonemeeting.restapi.tests.base import BaseTestCase
from plonemeeting.restapi.tests.config import base64_pdf_data
from plonemeeting.restapi.utils import IN_NAME_OF_UNAUTHORIZED
from plonemeeting.restapi.utils import IN_NAME_OF_USER_NOT_FOUND
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.tests.PloneMeetingTestCase import DEFAULT_USER_PASSWORD
from Products.PloneMeeting.utils import get_annexes

import pytz
import transaction


class testServiceAdd(BaseTestCase):
    """@item/@meeting POST endpoints."""

    layer = PM_REST_TEST_ADD_PROFILE_FUNCTIONAL

    def tearDown(self):
        self.api_session.close()
        transaction.abort()

    def test_restapi_add_item_config_id_not_found(self):
        """The 'config_id' parameter must be given"""
        transaction.begin()
        endpoint_url = "{0}/@item".format(self.portal_url)
        response = self.api_session.post(endpoint_url, json={"config_id": "unknown"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {u"message": CONFIG_ID_NOT_FOUND_ERROR % "unknown", u"type": u"BadRequest"},
        )

    def test_restapi_add_item_required_params(self):
        """A valid 'config_id' must be given"""
        self.changeUser("pmManager")
        pmFolder = self.getMeetingFolder()
        self.assertEqual(len(pmFolder.objectIds("MeetingItem")), 0)
        endpoint_url = "{0}/@item".format(self.portal_url)
        response = self.api_session.post(endpoint_url)
        transaction.commit()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {u"message": CONFIG_ID_ERROR, u"type": u"BadRequest"}
        )
        response = self.api_session.post(
            endpoint_url,
            json={
                "config_id": self.meetingConfig.getId(),
                "proposingGroup": self.developers_uid,
                "title": "My item",
            },
        )
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        self.assertEqual(len(pmFolder.objectIds("MeetingItem")), 1)
        item = pmFolder.get("my-item")
        self.assertEqual(item.Title(), "My item")
        self.assertEqual(item.getProposingGroup(), self.developers_uid)

    def test_restapi_add_item_ids_into_uids_error(self):
        """When creating an item, for convenience, fields holding organizations
           like "proposingGroup", "groupsInCharge", ... may have the organization id
           even if finally the UID is stored.
           If a wrong id/UID is given, an error is raised."""
        transaction.begin()
        endpoint_url = "{0}/@item".format(self.portal_url)
        response = self.api_session.post(
            endpoint_url,
            json={
                "config_id": self.meetingConfig.getId(),
                "proposingGroup": "wrong-id",
                "title": "My item",
            },
        )
        transaction.begin()
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u"message": ORG_FIELD_VALUE_ERROR % ("wrong-id", "proposingGroup"),
             u"type": u"BadRequest"}
        )

    def test_restapi_add_item_optional_fields(self):
        """When creating an item, given optional fields must be enabled in config."""
        cfg = self.meetingConfig
        self.assertFalse("category" in cfg.getUsedItemAttributes())
        self.assertFalse("notes" in cfg.getUsedItemAttributes())
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers_uid,
            "title": "My item",
            "category": "development",
            "notes": u"<p>My notes</p>",
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {u"message": OPTIONAL_FIELD_ERROR % "category", u"type": u"BadRequest"}
        )
        self._enableField('category')
        transaction.commit()
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {u"message": OPTIONAL_FIELD_ERROR % "notes", u"type": u"BadRequest"}
        )
        self._enableField("notes")
        transaction.commit()
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.Title(), json["title"])
        self.assertEqual(item.getProposingGroup(), json["proposingGroup"])
        self.assertEqual(item.getNotes(), json["notes"])

    def test_restapi_add_item_optional_fields_ignore_not_used_data(self):
        """When creating an item, given optional fields must be enabled in config,
           but when using parameter "ignore_not_used_data=true" then a warning
           is added instead raising an error."""
        cfg = self.meetingConfig
        self.assertFalse("notes" in cfg.getUsedItemAttributes())
        self.assertFalse("category" in cfg.getUsedItemAttributes())
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers_uid,
            "title": "My item",
            "category": "development",
            "notes": u"<p>My notes</p>",
            "ignore_not_used_data": True
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        self.assertTrue(OPTIONAL_FIELDS_WARNING % "category, notes" in response.json()['@warnings'])
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.Title(), json["title"])
        self.assertEqual(item.getProposingGroup(), json["proposingGroup"])
        # optional field not enabled was ignored
        self.assertFalse(item.getNotes())

    def test_restapi_add_item_org_fields(self):
        """When creating an item, values for fields storing organization
           uids may receive organizations ids instead to ease use by externals."""
        cfg = self.meetingConfig
        cfg.setOrderedGroupsInCharge([self.vendors_uid])
        self._enableField("groupsInCharge")
        self._enableField("associatedGroups")
        self.changeUser("pmManager")
        transaction.commit()
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "groupsInCharge": [self.vendors.getId(), self.developers.getId()],
            # use internal and external organizations by id and UID
            "associatedGroups": [self.vendors.getId(),
                                 self.developers_uid,
                                 "../{0}".format(self.orgOutside2.getId()),
                                 self.orgOutside1_uid, ],
            "optionalAdvisers": [self.vendors.getId()],
            "title": "My item",
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.Title(), json["title"])
        # turned to UIDs
        self.assertEqual(item.getProposingGroup(), self.developers_uid)
        self.assertEqual(
            item.getGroupsInCharge(), [self.vendors_uid, self.developers_uid]
        )
        self.assertEqual(
            item.getAssociatedGroups(),
            (self.vendors_uid, self.developers_uid, self.orgOutside2_uid, self.orgOutside1_uid)
        )
        self.assertEqual(item.getOptionalAdvisers(), (self.vendors_uid,))
        self.assertTrue(self.vendors_uid in item.adviceIndex)

    def test_restapi_add_item_can_not_use_org_outside_own_org(self):
        """Can not use an external org UID in relevant fields, excepted for the
           associatedGroups field (tested in test_restapi_add_item_org_fields)."""
        cfg = self.meetingConfig
        cfg.setOrderedGroupsInCharge([self.vendors_uid])
        self._enableField("groupsInCharge")
        self._enableField("associatedGroups")
        self.changeUser("pmManager")
        transaction.commit()
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "preferredMeeting": "unknown_uid",
            "optionalAdvisers": [self.vendors.getId(), self.orgOutside1_uid],
            "groupsInCharge": [self.vendors.getId(), self.orgOutside1_uid],
            "title": "My item",
        }
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u'message': u'[{\'field\': \'groupsInCharge\', \'message\': '
             u'u"Values [\'%s\'] are not allowed for vocabulary of element Groups in charge.", '
             u'\'error\': \'ValidationError\'}, {\'field\': \'optionalAdvisers\', '
             u'\'message\': u"Values [\'%s\'] are not allowed for vocabulary of element Optional advisers.", '
             u'\'error\': \'ValidationError\'}, {\'field\': \'preferredMeeting\', \'message\': '
             u'u"Values [\'unknown_uid\'] are not allowed for vocabulary of element Preferred meeting.", '
             u'\'error\': \'ValidationError\'}]' % (self.orgOutside1_uid, self.orgOutside1_uid),
             u'type': u'BadRequest'})
        # if preferredMeeting set to None, then created item will use ITEM_NO_PREFERRED_MEETING_VALUE
        json["preferredMeeting"] = None
        # check for groupsInCharge
        json["optionalAdvisers"] = [self.vendors.getId()]
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u'message': u'[{\'field\': \'groupsInCharge\', \'message\': u"Values [\'%s\'] '
             u'are not allowed for vocabulary of element Groups in charge.", '
             u'\'error\': \'ValidationError\'}]' % self.orgOutside1_uid,
             u'type': u'BadRequest'})
        # fix everything
        json["groupsInCharge"] = [self.vendors.getId()]
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 201, response.content)
        self.assertEqual(response.json()["preferredMeeting"],
                         {u'token': u'whatever', u'title': u'No preference'})

    def test_restapi_add_item_wf_transitions_present(self):
        """When creating an item, we may define "wf_transitions"
           until "present", in this case, item is "presented" in next meeting."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        # meeting in the future
        date = datetime.now() + timedelta(days=1)
        meeting = self.create("Meeting", date=date)
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "wf_transitions": ["propose", "validate", "present"]
        }
        transaction.commit()
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.query_state(), "presented")
        self.assertEqual(item.getMeeting(), meeting)

    def test_restapi_add_item_in_name_of(self):
        """Test while using 'in_name_of' parameter"""
        cfg = self.meetingConfig
        # must be (Meeting)Manager to use in_name_of
        self.api_session.auth = ("pmCreator1", DEFAULT_USER_PASSWORD)
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.vendors_uid,
            "title": "My item",
            "in_name_of": "pmCreator2",
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 401, response.content)
        self.assertEqual(
            response.json(),
            {
                u"message": IN_NAME_OF_UNAUTHORIZED % ("pmCreator1", "pmCreator2"),
                u"type": u"Unauthorized",
            },
        )
        # user must exist
        self.api_session.auth = ("pmManager", DEFAULT_USER_PASSWORD)
        json["in_name_of"] = "unknown"
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(
            response.json(),
            {u"message": IN_NAME_OF_USER_NOT_FOUND % "unknown", u"type": u"BadRequest"},
        )
        # now as MeetingManager for pmCreator2
        json["in_name_of"] = "pmCreator2"
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        # item was created in the pmCreator2 folder
        response_json = response.json()
        self.assertEqual(response_json["creators"], [u"pmCreator2"])
        self.assertTrue("Members/pmCreator2/mymeetings" in response_json["@id"])

    def test_restapi_add_item_external_identifier(self):
        """When creating an item, we may receive an externalIdentifier",
           in this case it is stored on the created item."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            # may sometimes be given as an integer
            "externalIdentifier": 123
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        # is forced to str when stored
        self.assertEqual(item.externalIdentifier, "123")
        brains = self.catalog(externalIdentifier="123")
        self.assertEqual(len(brains), 1)
        self.assertEqual(brains[0].UID, item.UID())

    def test_restapi_add_item_internal_notes(self):
        """Make sure we can create an item with internalNotes, as this field
           is protected by specific permission/role."""
        cfg = self.meetingConfig
        self._enableField("internalNotes")
        self._activate_config('itemInternalNotesEditableBy',
                              'suffix_proposing_group_creators',
                              keep_existing=False)
        transaction.commit()
        self.changeUser("pmCreator1")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "internalNotes": u"<p>My internal notes.</p>"
        }
        self.api_session.auth = ("pmCreator1", DEFAULT_USER_PASSWORD)
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 201, response.content)
        self.assertEqual(
            response.json()["internalNotes"],
            {u'data': u"<p>My internal notes.</p>",
             u'content-type': u'text/html'})
        # as we need proposingGroup in data when using internalNotes
        # check that the code handles proposingGroup not being in data
        json.pop("proposingGroup")
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {
                u'message':
                    u"[{'field': 'proposingGroup', 'message': u'A proposing group is required.', "
                    u"'error': 'ValidationError'}]",
                u'type':
                    u'BadRequest'}
        )

    def test_restapi_add_item_ignore_validation_for(self):
        """When creating an item, it is possible to define
           a list of fields to bypass validation for if it is empty."""
        cfg = self.meetingConfig
        self._enableField('category')
        self._enableField("classifier")
        transaction.commit()
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "category": "unknown",
            "classifier": "classifier1"
        }
        # when using categories, creating an item without category fails
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u'message': u"[{'field': 'category', 'message': u'Please select a category.', "
                         u"'error': 'ValidationError'}]",
             u'type': u'BadRequest'}
        )
        # when using "ignore_validation_for", some required fields
        # validation may not be bypassed
        json["title"] = ""
        json["ignore_validation_for"] = ["title"]
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {
                u"message":
                    IGNORE_VALIDATION_FOR_REQUIRED_ERROR % "title, proposingGroup",
                u"type": u"BadRequest",
            }
        )
        # when using "ignore_validation_for", only empty values
        # (or not given at all) validation may be bypassed
        json["title"] = "My item"
        json["ignore_validation_for"] = ["category"]
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {
                u"message": IGNORE_VALIDATION_FOR_VALUED_ERROR % "category",
                u"type": u"BadRequest",
            }
        )
        # use "ignore_validation_for" correctly
        # bypass category validation and classifier validation
        # pass empty classifier/category
        # define a ignore_validation_for a field that it not in json data
        # check also that such an item may be set to WF state "validated"
        json["ignore_validation_for"] = ["category", "classifier", "groupsInCharge"]
        json["category"] = None
        json["classifier"] = None
        json["wf_transitions"] = ["propose", "validate"]
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.getCategory(), "")
        self.assertEqual(item.getClassifier(), "")
        # a warning was added nevertheless
        self.assertEqual(response.json()['@warnings'],
                         [IGNORE_VALIDATION_FOR_WARNING %
                          "category, classifier, groupsInCharge"])
        self.assertEqual(item.query_state(), "validated")

    def test_restapi_add_item_wf_transitions(self):
        """When creating an item, we may define "wf_transitions"."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "wf_transitions": ["propose", "validate"]
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.query_state(), "validated")

    def test_restapi_add_annex_to_existing_element(self):
        """Use the @annex POST endpoint to create an annex."""
        cfg = self.meetingConfig
        self._removeConfigObjectsFor(cfg)
        self.changeUser("pmManager")
        item = self.create('MeetingItem')
        item_uid = item.UID()
        meeting = self.create('Meeting')
        meeting_uid = meeting.UID()
        transaction.commit()

        # add annex to item
        json = {
            "title": "My annex",
            "content_category": "wrong-annex",
            "file": {
                "data": "123456",
                "encoding": "ascii",
                "filename": "file.txt"}
        }
        endpoint_url = "{0}/@annex/{1}".format(self.portal_url, item_uid)
        # wrong content_category
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u"message": ANNEX_CONTENT_CATEGORY_ERROR % "wrong-annex",
             u"type": u"BadRequest"}
        )
        # add annex to item correctly
        json["content_category"] = "item-annex"
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        # adding an annex without "content_category" will use default one
        # the default annex type is "financial-analysis"
        json.pop("content_category")
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 201, response.content)
        self.assertEqual(response.json()["content_category"]["title"], u'Financial analysis')
        # add annexDecision to item correctly
        json["content_category"] = "decision-annex"
        json["decision_related"] = True
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 201, response.content)
        # add annex to meeting
        # can not use parameter "decision_related" on a meeting
        json["content_category"] = "meeting-annex"
        endpoint_url = endpoint_url.replace(item_uid, meeting_uid)
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u"message": ANNEX_DECISION_RELATED_NOT_ITEM_ERROR,
             u"type": u"BadRequest"}
        )
        # add annex to meeting correctly
        json["decision_related"] = False
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 201, response.content)
        transaction.begin()

        # annexes were added to item and meeting
        item_annexes = get_annexes(item, ["annex"])
        self.assertEqual(len(item_annexes), 2)
        decision_annexes = get_annexes(item, ["annexDecision"])
        self.assertEqual(len(decision_annexes), 1)
        meeting_annexes = get_annexes(meeting)
        self.assertEqual(len(meeting_annexes), 1)

    def test_restapi_add_item_clean_html(self):
        """When creating an item, HTML will be cleaned by default.
           Adding 'clean_html':false in the body will disable it."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        dirty_html = '<span class="ms-class">Hello, &#xa0; la d\xc3\xa9cision ' \
            '\xc3\xa9tait longue!</span><br /><strong>'
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            # some dirty HTML
            "decision": dirty_html,
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item1 = pmFolder.objectValues()[-1]
        # decision was cleaned
        self.assertEqual(
            item1.getDecision(),
            '<p><span class="ms-class">Hello, \xc2\xa0 la d\xc3\xa9cision '
            '\xc3\xa9tait longue!</span><br /><strong></strong></p>')
        # create item with clean_html=False
        json['clean_html'] = False
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        # decision was not cleaned
        item2 = pmFolder.objectValues()[-1]
        self.assertEqual(item2.getDecision(), dirty_html)

    def test_restapi_add_clean_meeting(self):
        """When creating an meeting, HTML will be cleaned by default."""
        transaction.begin()
        self._enableField("observations", related_to="Meeting")
        # make sure creating a meeting work when using attendees
        self._enableField("attendees", related_to="Meeting")
        transaction.commit()
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@meeting".format(self.portal_url)
        dirty_html = '<span class="ms-class">&#xa0; hello h\xc3\xa9h\xc3\xa9'
        date = datetime.now(tz=pytz.UTC).isoformat().replace("+00:00", "Z")
        json = {
            "config_id": cfg.getId(),
            "date": date,
            # some dirty HTML
            "observations": dirty_html,
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        meeting = object_values(pmFolder, "Meeting")[-1]
        # dirty_html was cleaned
        self.assertEqual(
            meeting.observations.raw,
            u'<p><span class="ms-class">\xa0 hello h\xe9h\xe9</span></p>')
        # this fails with AT Meeting because mimetype is considered text/plain
        # create meeting with clean_html=False
        json['clean_html'] = False
        # change date, can not create several meeting with same date
        date = date[0:3] + str(int(date[3]) + 1) + date[4:]
        json['date'] = date
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        meeting2 = object_values(pmFolder, "Meeting")[-1]
        # the "&#xa0;" is replaced by "\xa0" by Plone but could also
        # be removed while using appy.pod pre processor
        meeting_dirty_html = safe_unicode(dirty_html).replace(u'&#xa0;', u'\xa0')
        self.assertEqual(meeting2.observations.raw, meeting_dirty_html)
        # trying to add meeting with same date will fail
        response = self.api_session.post(endpoint_url, json=json)
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u'message': u"[{'message': 'A meeting having the same date and hour "
                u"already exists. Please choose another date and/or hour.', "
                u"'error': 'ValidationError'}]",
             u'type': u'BadRequest'})

    def test_restapi_add_item_can_not_create_empty(self):
        """Test that an empty item can not be created."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {"config_id": cfg.getId()}
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {u'message':
                u"[{'field': 'proposingGroup', 'message': u'A proposing group is required.', "
                u"'error': 'ValidationError'}, {'field': 'title', 'message': u'Purpose is required, "
                u"please correct.', 'error': 'ValidationError'}]", u'type': u'BadRequest'}
        )

    def test_restapi_add_item_manually_linked_items(self):
        """Create an item and link it to another item using MeetingItem.manuallyLinkedItems."""
        cfg = self.meetingConfig
        self._enableField("manuallyLinkedItems")
        transaction.commit()
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "Item 1"
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item1 = pmFolder.objectValues()[-1]
        # create a second item
        json["title"] = "Item 2"
        json["manuallyLinkedItems"] = [item1.UID()]
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        item2 = pmFolder.objectValues()[-1]
        self.assertEqual(item1.getManuallyLinkedItems(), [item2])
        self.assertEqual(item2.getManuallyLinkedItems(), [item1])
        # create a third item
        json["title"] = "Item 3"
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        item3 = pmFolder.objectValues()[-1]
        self.assertEqual(item1.getManuallyLinkedItems(), [item2, item3])
        self.assertEqual(item2.getManuallyLinkedItems(), [item1, item3])
        self.assertEqual(item3.getManuallyLinkedItems(), [item1, item2])


class testServiceAddWithAnnexes(BaseTestCase):
    """@item/@meeting POST endpoints with annexes."""

    layer = PM_REST_TEST_ADD_ANNEXES_PROFILE_FUNCTIONAL

    def tearDown(self):
        self.api_session.close()
        transaction.abort()

    def test_restapi_add_item_with_annexes(self):
        """When creating an item, we may add annexes as __children__."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex",
                    "content_category": "item-annex",
                    "file": {
                        "data": "123456",
                        "encoding": "ascii",
                        "filename": "file.txt",
                    },
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.Title(), json["title"])
        annex = get_annexes(item)[0]
        self.assertEqual(annex.title, json["__children__"][0]["title"])
        self.assertEqual(
            annex.content_category,
            calculate_category_id(cfg.annexes_types.item_annexes.get("item-annex")),
        )
        self.assertEqual(
            annex.file.filename, json["__children__"][0]["file"]["filename"]
        )
        self.assertEqual(annex.file.size, 6)
        self.assertEqual(annex.file.contentType, "text/plain")

    def test_restapi_add_item_with_annexes_and_check_encoding(self):
        """When creating an item, we may add annexes as __children__,
           if encoding not given, we assume it is 'base64'."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex",
                    "content_category": "item-annex",
                    "file": {"data": base64_pdf_data, "filename": "file.pdf"},
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        annex = get_annexes(item)[0]
        self.assertEqual(
            annex.file.filename, json["__children__"][0]["file"]["filename"]
        )
        self.assertEqual(annex.file.size, 6475)
        self.assertEqual(annex.file.contentType, "application/pdf")

    def test_restapi_add_item_with_annexes_and_content_category(self):
        """When creating an item, we may add annexes as __children__,
           if using wrong content_category, cration is aborted."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex",
                    "content_category": "unknown",
                    "file": {
                        "data": "123456",
                        "encoding": "ascii",
                        "filename": "file.txt",
                    },
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {
                u"message": ANNEX_CONTENT_CATEGORY_ERROR % "unknown",
                u"type": u"BadRequest",
            }
        )

    def test_restapi_add_item_with_annexes_and_filename_or_content_type_required(self):
        """When creating an item, we may add annexes as __children__,
           one of 'filename' or 'content-type' is required to determinate content_type."""
        transaction.begin()
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex",
                    "content_category": "unknown",
                    "file": {
                        "data": "123456",
                        "encoding": "ascii",
                        "filename": "file.txt",
                    },
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 400, response.content)
        self.assertEqual(
            response.json(),
            {
                u"message": ANNEX_CONTENT_CATEGORY_ERROR % "unknown",
                u"type": u"BadRequest",
            }
        )

    def test_restapi_add_item_with_annexes_children(self):
        """When creating an item, we may add annexes as __children__,
           we may add several annexes at once."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex 1",
                    "content_category": "item-annex",
                    "file": {
                        "data": "123456",
                        "encoding": "ascii",
                        "filename": "file.txt",
                    },
                },
                {
                    "@type": "annex",
                    "title": "My annex 2",
                    "content_category": "item-annex",
                    "file": {"data": base64_pdf_data, "filename": "file.pdf"},
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.commit()
        # this tries to manage randomly failing test
        try:
            self.assertEqual(response.status_code, 201, response.content)
        except Exception:
            response = self.api_session.post(endpoint_url, json=json)
            transaction.commit()
            self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        annex1 = get_annexes(item)[0]
        annex2 = get_annexes(item)[1]
        self.assertEqual(
            annex1.file.filename, json["__children__"][0]["file"]["filename"]
        )
        self.assertEqual(annex1.file.size, 6)
        self.assertEqual(annex1.file.contentType, "text/plain")
        self.assertEqual(
            annex2.file.filename, json["__children__"][1]["file"]["filename"]
        )
        self.assertEqual(annex2.file.size, 6475)
        self.assertEqual(annex2.file.contentType, "application/pdf")

    def test_restapi_add_item_with_internal_number(self):
        """When creating an item and using internal_number, it is correctly incremented."""
        cfg = self.meetingConfig
        set_settings({cfg.getItemTypeName(): {'u': False, 'nb': 5, 'expr': u'number'}})
        transaction.commit()
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
        }
        self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.internal_number, 5)
        self.assertEqual(self.catalog(internal_number=5)[0].UID, item.UID())
        # returned when using metadata_fields even if not a real field
        # internal_number, not a real field, managed manually
        endpoint_url = "{0}/@get?uid={1}" \
            "&metadata_fields=internal_number".format(self.portal_url, item.UID())
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json["internal_number"], 5)
        # also returned when using fullobjects
        endpoint_url = endpoint_url.replace("&metadata_fields=internal_number", "&fullobjects")
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json["internal_number"], 5)

    def test_restapi_add_a_meeting_with_annexes(self):
        """When creating a meeting, we may add annexes as __children__,
           we may add several annexes at once."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex 1",
                    "file": {
                        "data": "123456",
                        "encoding": "ascii",
                        "filename": "file.txt",
                    },
                },
                {
                    "@type": "annex",
                    "title": "My annex 2",
                    "file": {"data": base64_pdf_data, "filename": "file.pdf"},
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        pmFolder = self.getMeetingFolder()
        meeting = pmFolder.objectValues()[-1]
        annex1 = get_annexes(meeting)[0]
        annex2 = get_annexes(meeting)[1]
        self.assertEqual(
            annex1.file.filename, json["__children__"][0]["file"]["filename"]
        )
        self.assertEqual(annex1.file.size, 6)
        self.assertEqual(annex1.file.contentType, "text/plain")
        self.assertEqual(
            annex2.file.filename, json["__children__"][1]["file"]["filename"]
        )
        self.assertEqual(annex2.file.size, 6475)
        self.assertEqual(annex2.file.contentType, "application/pdf")

    def test_restapi_add_item_with_annexes_and_in_name_of(self):
        """When creating an item, we may add annexes as __children__ using `in_name_of`
        parameter"""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = "{0}/@item".format(self.portal_url)
        json = {
            "config_id": cfg.getId(),
            "proposingGroup": self.developers.getId(),
            "title": "My item",
            "in_name_of": "pmCreator1",
            "__children__": [
                {
                    "@type": "annex",
                    "title": "My annex",
                    "content_category": "item-annex",
                    "file": {
                        "data": "123456",
                        "encoding": "ascii",
                        "filename": "file.txt",
                    },
                },
            ],
        }
        response = self.api_session.post(endpoint_url, json=json)
        transaction.begin()
        self.assertEqual(response.status_code, 201, response.content)
        self.changeUser("pmCreator1")
        pmFolder = self.getMeetingFolder()
        item = pmFolder.objectValues()[-1]
        self.assertEqual(item.Title(), json["title"])
        annex = get_annexes(item)[0]
        self.assertEqual(annex.title, json["__children__"][0]["title"])
        self.assertEqual(
            annex.content_category,
            calculate_category_id(cfg.annexes_types.item_annexes.get("item-annex")),
        )
        self.assertEqual(
            annex.file.filename, json["__children__"][0]["file"]["filename"]
        )
        self.assertEqual(annex.file.size, 6)
        self.assertEqual(annex.file.contentType, "text/plain")


def test_suite():
    from unittest import TestSuite, makeSuite

    suite = TestSuite()
    # add a prefix to avoid every PM tests to be executed
    suite.addTest(makeSuite(testServiceAdd, prefix="test_restapi_"))
    suite.addTest(makeSuite(testServiceAddWithAnnexes, prefix="test_restapi_"))
    return suite



# -*- coding: utf-8 -*-

from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plonemeeting.restapi.serializer.base import BaseDXSerializeFolderToJson
from plonemeeting.restapi.serializer.summary import PMBrainJSONSummarySerializer
from Products.PloneMeeting.content.held_position import IPMHeldPosition
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


class SerializeHeldPositionToJsonBase(object):
    """ """

    def _available_extra_includes(self, result):
        """ """
        result["@extra_includes"] = ["person"]
        return result

    def _extra_include(self, result):
        """ """
        extra_include = self._get_asked_extra_include()
        if "person" in extra_include:
            person = self.context.get_person()
            serializer = self._get_serializer(person, "person")
            result["extra_include_person"] = serializer()
        return result


@implementer(ISerializeToJson)
@adapter(IPMHeldPosition, Interface)
class SerializeToJson(SerializeHeldPositionToJsonBase, BaseDXSerializeFolderToJson):
    """ """


@implementer(ISerializeToJsonSummary)
@adapter(IPMHeldPosition, Interface)
class SerializeToJsonSummary(SerializeHeldPositionToJsonBase, PMBrainJSONSummarySerializer):
    """ """



# -*- coding: utf-8 -*-

# XXX Override to query the serializer adapter with the real object
# XXX instead the catalog brain as we register different summary serializers.
# 2 things changed :
# - use LazyMap instead Lazy to be more specific
# - call summary serializer on object instead brain, check the XXX here under

from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.catalog import LazyCatalogResultSerializer
from plonemeeting.restapi import logger
from plonemeeting.restapi.utils import use_obj_serializer
from Products.ZCatalog.Lazy import LazyMap
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface


log = logger


@implementer(ISerializeToJson)
# XXX register for LazyMap instead Lazy
@adapter(LazyMap, Interface)
class PMLazyCatalogResultSerializer(LazyCatalogResultSerializer):
    """ """

    def _use_obj_serializer(self, fullobjects):
        """Check if we need to use our own implementation of the summary serializer,
           necessary to manage additional_values and extra_include."""
        # if we have any parameter that ask for something else in the result
        # then we use our serializer, the summary serializer will only be used
        # to receive basic summary informations
        return fullobjects or use_obj_serializer(self.request.form)

    def __call__(self, fullobjects=False):
        batch = HypermediaBatch(self.request, self.lazy_resultset)

        results = {}
        results["@id"] = batch.canonical_url
        results["items_total"] = batch.items_total
        links = batch.links
        if links:
            results["batching"] = links

        results["items"] = []
        use_obj_serializer = self._use_obj_serializer(fullobjects)
        for brain in batch:
            try:
                obj = use_obj_serializer and brain.getObject() or brain
            except AttributeError:
                # Guard in case the brain returned refers to an object that doesn't
                # exists because it failed to uncatalog itself or the catalog has
                # stale cataloged objects for some reason
                log.warning(
                    "Brain getObject error: {} doesn't exist anymore".format(
                        brain.getPath()
                    )
                )
                continue

            if use_obj_serializer:
                result = getMultiAdapter(
                    (obj, self.request), ISerializeToJson
                )(include_items=False)
            else:
                result = getMultiAdapter(
                    (obj, self.request), ISerializeToJsonSummary
                )()

            results["items"].append(result)

        return results



# -*- coding: utf-8 -*-

from Acquisition import aq_inner
from Acquisition import aq_parent
from collective.documentgenerator.interfaces import IGenerablePODTemplates
from collective.iconifiedcategory.utils import _categorized_elements
from collective.iconifiedcategory.utils import get_categorized_elements
from imio.helpers.content import base_hasattr
from imio.helpers.content import get_vocab
from imio.restapi.serializer.base import SerializeFolderToJson as IMIODXSerializeFolderToJson
from imio.restapi.serializer.base import SerializeToJson as IMIODXSerializeToJson
from imio.restapi.utils import serialize_term
from plone import api
from plone.autoform.interfaces import READ_PERMISSIONS_KEY
from plone.dexterity.interfaces import IDexterityContainer
from plone.dexterity.interfaces import IDexterityContent
from plone.dexterity.utils import iterSchemata
from plone.restapi.batching import HypermediaBatch
from plone.restapi.deserializer import boolean_value
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.interfaces import IObjectPrimaryFieldTarget
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.atcontent import SerializeFolderToJson as ATSerializeFolderToJson
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.serializer.expansion import expandable_elements
from plone.restapi.serializer.nextprev import NextPrevious
from plone.supermodel.utils import mergedTaggedValueDict
from plonemeeting.restapi import logger
from plonemeeting.restapi.config import ANNEXES_FILTER_VALUES
from plonemeeting.restapi.interfaces import IPMRestapiLayer
from plonemeeting.restapi.utils import get_param
from plonemeeting.restapi.utils import get_serializer
from zope.component import adapter
from zope.component import ComponentLookupError
from zope.component import getAdapter
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface
from zope.schema import getFields


NO_VOCAB_ERROR = "Trying to get vocabulary for field %s but it is None"
SUFFIXED_CHOICES_PATTERN = "%s__choices"


def serialize_pod_templates(context, serializer):
    """Serialize generable POD templates for p_context."""
    # get generatable POD template for self.context
    adapter = getAdapter(context, IGenerablePODTemplates)
    generable_templates = adapter.get_generable_templates()
    result = []
    for pod_template in generable_templates:
        pod_serializer = serializer._get_serializer(pod_template, "pod_templates")
        # to be used to compute url to generate element
        pod_serializer.original_context = context
        result.append(pod_serializer())
    return result


def _get_annex_uids(data, filters):
    """Compatibility for PM4.1.x/4.2.x, as filters can not be passed to
       get_categorized_elements, we manage it manually.
       So we compute uids to keep and we pass it to get_categorized_elements.
       XXX to be removed when using 4.2.x+ for everybody."""
    uids = []
    for annex_uid, annex_infos in data.items():
        # check filters
        keep = True
        for k, v in filters.items():
            if annex_infos[k] != v:
                keep = False
                break
        if keep:
            uids.append(annex_uid)
    return uids


def serialize_annexes(context, filters, extra_include_name=None, base_serializer=None):
    """Serialize visible annexes regarding filters found in request."""
    result = []
    # can not use parameter filters due to compatibility 4.1.x, use uids instead
    # XXX to be adapted, just pass filters to get_categorized_elements and
    # remove uids computation
    uids = _get_annex_uids(_categorized_elements(context), filters)
    # uids must contains something or passing an empty list means do not filter on uids
    uids = uids or ["dummy_uid"]
    annexes = get_categorized_elements(
        context, result_type="objects", uids=uids)  # , filters=filters)
    for annex in annexes:
        serializer = get_serializer(
            annex, extra_include_name=extra_include_name, serializer=base_serializer)
        result.append(serializer())
    return result


def serialize_attendees(context, attendee=None, extra_include_name=None, base_serializer=None):
    """ """
    result = []
    tool = api.portal.get_tool("portal_plonemeeting")
    cfg = tool.getMeetingConfig(context)
    attendee_types = {}
    is_meeting = context.__class__.__name__ == "Meeting"
    meeting = context if is_meeting else context.getMeeting()
    # return empty result when called on item out of meeting
    if meeting is None:
        return result
    if is_meeting:
        attendee_types.update({attendee_uid: 'present' for attendee_uid in context.get_attendees()})
        attendee_types.update({absent_uid: 'absent' for absent_uid in context.get_absents()})
        attendee_types.update({excused_uid: 'excused' for excused_uid in context.get_excused()})
    else:
        # MeetingItem
        attendee_types.update({attendee_uid: 'present' for attendee_uid in context.get_attendees()})
        attendee_types.update({absent_uid: 'absent' for absent_uid in
                               meeting.get_absents() + context.get_item_absents()})
        attendee_types.update({excused_uid: 'excused' for excused_uid in
                               meeting.get_excused() + context.get_item_excused()})

    # initialize voter
    voters = meeting.get_voters()
    signatories = context.get_signatories(include_position_type=True) \
        if is_meeting else context.get_item_signatories(include_position_type=True)
    non_attendees = context.get_item_non_attendees()
    attendees = [attendee] if attendee is not None else context.get_all_attendees(the_objects=True)
    for attendee in attendees:
        serializer = get_serializer(
            attendee,
            extra_include_name=extra_include_name,
            serializer=base_serializer,
            interface=ISerializeToJson)
        # for is_meeting position_type, let the serializer manage it
        serializer._init()
        serializer.fullobjects = False
        if is_meeting:
            serializer.metadata_fields = ['position_type']
        serialized = serializer()
        # 'UID' could not be in serialized if include_base_data=false
        serialized_uid = serialized.get('UID', attendee.UID())
        # for not is_meeting position_type, manage it manually as it may be redefined
        # attendee (context) is used to return correct value depending on gender/number
        position_type_vocab = None
        if not is_meeting:
            position_type_vocab = get_vocab(attendee, "PMPositionTypes")
            serialized['position_type'] = serialize_term(
                meeting.get_attendee_position_for(context.UID(), serialized_uid),
                position_type_vocab)
        # manage "attendee_type"
        attendee_type = "non_attendee" if \
            serialized_uid in non_attendees else attendee_types[serialized_uid]
        serialized["attendee_type"] = {
            'token': attendee_type,
            'title': translate(attendee_type, domain="PloneMeeting", context=context.REQUEST)}
        # manage "signatory"
        serialized["signatory"] = None
        serialized["signatory_position_type"] = None
        signatory_infos = signatories.get(serialized_uid, {})
        if signatory_infos:
            position_type_vocab = position_type_vocab if position_type_vocab is not None \
                else get_vocab(attendee, "PMPositionTypes")
            serialized["signatory"] = signatory_infos['signature_number']
            serialized["signatory_position_type"] = serialize_term(
                signatory_infos['position_type'],
                position_type_vocab)
        # manage "voter"
        serialized["voter"] = serialized_uid in voters
        result.append(serialized)
        # manage hp title that could change on item
        item = None
        if not is_meeting:
            item = context
        serialized["title"] = meeting.get_attendee_short_title(attendee, cfg, item=item)
    return result


def serialize_extra_include_annexes(result, serializer):
    """ """
    # compute filters to get annexes
    filters = {}
    for filter_value in ANNEXES_FILTER_VALUES:
        value = serializer.get_param(
            filter_value,
            default=None,
            extra_include_name="annexes")
        if value is not None:
            filters[filter_value] = boolean_value(value)
    result["extra_include_annexes"] = serialize_annexes(
        serializer.context, filters, "annexes", serializer)
    return result


class BaseSerializeToJson(object):
    """__call__ must be redefined by class heritating from BaseSerializeToJson."""

    def _init(self):
        """ """
        # if already _init, pass, could have been changed manually
        if base_hasattr(self, 'metadata_fields'):
            return
        self.metadata_fields = self.get_param('metadata_fields', [])
        self.asked_extra_include = self._get_asked_extra_include()
        self.asked_additional_values = self._get_asked_additional_values()
        self.asked_includes = self._get_asked_includes()
        # fullobjects is True:
        # if actually asked in request
        # or if nothing else asked
        self.fullobjects = boolean_value(self.get_param('fullobjects', False)) or \
            (not self.metadata_fields and
             not self.asked_extra_include and
             not self.asked_additional_values and
             not self.asked_includes)

    def _get_workflow_state(self, obj):
        wftool = api.portal.get_tool("portal_workflow")
        review_state = wftool.getInfoFor(obj, name="review_state", default=None)
        return review_state

    def _available_includes(self):
        """Internal function to determinate what are the available include_... values."""
        return [v.lstrip('_') for v in dir(self) if v.startswith('_include_')]

    def _available_extra_includes(self, result):
        """ """
        result["@extra_includes"] = []
        return result

    def _get_asked_extra_include(self):
        """ """
        extra_include = self.get_param("extra_include", [])
        # filter on _available_extra_includes this make we do not forget
        # to add an extra_include to _available_extra_includes
        aeis = self._available_extra_includes({})["@extra_includes"]
        filtered_extra_include = [
            ei for ei in extra_include if ei in aeis]
        # handle available extra_include starting with *
        starting_available_extra_include = tuple(
            aei[1:] for aei in aeis if aei.startswith('*'))
        # handle available extra_include ending with *
        ending_available_extra_include = tuple(
            aei[:-1] for aei in aeis if aei.endswith('*'))
        # handle available extra_include containing something
        containing_available_extra_include = tuple(
            aei[1:-1] for aei in aeis
            if aei.endswith('*') and aei.startswith('*'))
        if starting_available_extra_include or \
           ending_available_extra_include or \
           containing_available_extra_include:
            filtered_extra_include += [
                ei for ei in extra_include
                if ei.startswith(starting_available_extra_include) or
                ei.endswith(ending_available_extra_include) or
                [caei for caei in containing_available_extra_include if caei in ei]]
        return filtered_extra_include

    def _extra_include(self, result):
        """ """
        return result

    def _get_asked_additional_values(self):
        """ """
        # when fullobjects, additional_values default is ["*"]
        # otherwise additional_values default is []
        # self.fullobjects is still not computed
        if boolean_value(self.get_param("fullobjects", False)) is True:
            additional_values = self.get_param("additional_values", ["*"])
        else:
            additional_values = self.get_param("additional_values", [])
        return additional_values

    def _get_asked_includes(self):
        """ """
        return [v for v in self._available_includes()
                if self.get_param(v, None) is not None]

    def _additional_values(self, result, additional_values):
        """ """
        return result

    def _include_custom(self, obj, result):
        """Custom part made to be overrided when necessary."""
        return {}

    def _after__call__(self, obj, result):
        """ """
        # Include custom
        result.update(self._include_custom(obj, result))

        # only call _extra_include if relevant
        result = self._available_extra_includes(result)
        if self.get_param("extra_include", []):
            result = self._extra_include(result)

        # complete with additional values if relevant
        if self.asked_additional_values:
            result = self._additional_values(result, self.asked_additional_values)

        return result

    def _get_serializer(self, obj, extra_include_name):
        """ """
        return get_serializer(obj,
                              extra_include_name,
                              serializer=self)

    def get_param(self, value, default=False, extra_include_name=None):
        """ """
        return get_param(value,
                         default=default,
                         extra_include_name=extra_include_name,
                         serializer=self)


class ContentSerializeToJson(BaseSerializeToJson):
    """ """

    def _include_base_data(self, obj):
        """ """
        result = {}

        if self.get_param("include_base_data", True):
            result = {
                # '@context': 'http://www.w3.org/ns/hydra/context.jsonld',
                "@id": obj.absolute_url(),
                "id": obj.id,
                "@type": obj.portal_type,
                "created": json_compatible(obj.created()),
                "modified": json_compatible(obj.modified()),
                "review_state": self._get_workflow_state(obj),
                "UID": obj.UID(),
                "title": obj.Title(),
            }
            # add "enabled" if it exists
            if base_hasattr(obj, "enabled"):
                result["enabled"] = obj.enabled
        return result

    def _include_nextprev(self, obj):
        """ """
        result = {}
        if self.fullobjects or self.get_param('include_nextprev'):
            nextprevious = NextPrevious(obj)
            result.update(
                {"previous_item": nextprevious.previous, "next_item": nextprevious.next}
            )
        return result

    def _include_parent(self, obj):
        """ """
        result = {}
        # parent is only included when specifically asked, even when fullobjects
        if self.get_param('include_parent', False):
            parent = aq_parent(aq_inner(obj))
            parent_summary = getMultiAdapter(
                (parent, self.request), ISerializeToJsonSummary
            )()
            result["parent"] = parent_summary
        return result

    def _include_components(self, obj):
        """ """
        result = {}
        if self.fullobjects or self.get_param('include_components'):
            result.update(expandable_elements(self.context, self.request))
        return result

    def _include_fields(self, obj):
        """ """
        raise NotImplementedError

    def _include_choices_for(self, obj):
        """ """
        raise NotImplementedError

    def _include_items(self, obj, include_items):
        """ """
        result = {}
        include_items = self.get_param("include_items", include_items)
        if include_items:
            query = self._build_query()

            catalog = api.portal.get_tool("portal_catalog")
            brains = catalog(query)

            batch = HypermediaBatch(self.request, brains)

            result["items_total"] = batch.items_total
            if batch.links:
                result["batching"] = batch.links

            if self.fullobjects:
                result["items"] = getMultiAdapter(
                    (brains, self.request), ISerializeToJson
                )(fullobjects=True)["items"]
            else:
                result["items"] = [
                    getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()
                    for brain in batch
                ]
        return result

    def _include_target_url(self, obj):
        """ """
        result = {}
        if self.fullobjects or self.get_param('include_target_url'):
            target_url = None
            try:
                target_url = getMultiAdapter(
                    (self.context, self.request), IObjectPrimaryFieldTarget
                )()
            except ComponentLookupError:
                # only available on DX content
                pass
            if target_url:
                result["targetUrl"] = target_url
        return result

    def _include_allow_discussion(self, obj):
        """ """
        result = {}
        if self.fullobjects or self.get_param('include_allow_discussion'):
            result["allow_discussion"] = getMultiAdapter(
                (self.context, self.request), name="conversation_view"
            ).enabled()
        return result

    def _include_layout(self, obj):
        """ """
        result = {}
        if self.fullobjects or self.get_param('include_layout'):
            result["layout"] = self.context.getLayout()
        return result

    def getVersion(self, version):
        """ """
        return self.context

    def __call__(self, version=None, include_items=False):
        """Change include_items=False by default.
           Override to manage a way to get only what we want :
           - fullobjects, by default this will make element behave like in plone.restapi;
           - include_parent=False, to include the "parent" in the result;
           - include_nextprev=False, to include next/prev information;
           - include_components=True, to include @components (expandables);
           - metadata_fields=[], when given, this will only return selected values.
           """
        version = "current" if version is None else version
        obj = self.getVersion(version)
        self._init()

        # Include base_data
        result = self._include_base_data(obj)

        # Include parent
        result.update(self._include_parent(obj))

        # Include next/prev information
        result.update(self._include_nextprev(obj))

        # Include expandable elements (@components)
        result.update(self._include_components(obj))

        # Include metadata_fields
        result.update(self._include_fields(obj))

        # Include vocabulary values
        result.update(self._include_choices_for(obj))

        # Include items
        result.update(self._include_items(obj, include_items))

        # Include targetUrl
        result.update(self._include_target_url(obj))

        # Include allow_discussion
        result.update(self._include_allow_discussion(obj))

        # Include layout
        result.update(self._include_layout(obj))

        # call _after__call__ that manages additional_values and extra_includes
        result = self._after__call__(obj, result)
        return result


@implementer(ISerializeToJson)
@adapter(Interface, Interface)
class BaseATSerializeFolderToJson(ContentSerializeToJson, ATSerializeFolderToJson):
    """ """

    def _get_readable_fields(self, obj, field_names=[]):
        """ """
        fields = []
        for field in obj.Schema().fields():
            field_name = field.getName()
            if field_names and field_name not in field_names:
                continue
            if "r" not in field.mode or not field.checkPermission(
                "r", obj
            ):  # noqa: E501
                continue
            fields.append((field_name, field))
        return fields

    def _include_choices_for(self, obj):
        """ """
        result = {}
        choices_for = self.get_param('include_choices_for', [])
        if choices_for:
            for field_name, field in self._get_readable_fields(
                    obj, field_names=choices_for):
                if not field.vocabulary and not field.vocabulary_factory:
                    logger.warning(NO_VOCAB_ERROR % field_name)
                    continue
                vocab = field.Vocabulary(obj)
                data_name = SUFFIXED_CHOICES_PATTERN % field_name
                result[data_name] = []
                for term_id, term_title in vocab.items():
                    if term_id not in ['', '_none_']:
                        result[data_name].append(
                            {'token': term_id, 'title': term_title})
        return result

    def _include_fields(self, obj):
        """ """
        result = {}
        # Compute fields if fullobjects or metadata_fields
        if self.fullobjects or self.metadata_fields:
            for field_name, field in self._get_readable_fields(
                    obj, field_names=self.metadata_fields or []):
                serializer = queryMultiAdapter(
                    (field, self.context, self.request), IFieldSerializer
                )
                if serializer is not None:
                    result[field_name] = serializer()
        return result


@implementer(ISerializeToJson)
@adapter(IDexterityContent, IPMRestapiLayer)
class BaseDXSerializeToJson(ContentSerializeToJson, IMIODXSerializeToJson):
    """ """

    def _get_readable_fields(self, obj, field_names=[]):
        """ """
        fields = []
        for schema in iterSchemata(self.context):
            read_permissions = mergedTaggedValueDict(schema, READ_PERMISSIONS_KEY)
            for field_name, field in getFields(schema).items():
                if field_names and field_name not in field_names:
                    continue
                if not self.check_permission(read_permissions.get(field_name), obj):
                    continue
                fields.append((field_name, field))
        return fields

    def _include_choices_for(self, obj):
        """ """
        result = {}
        choices_for = self.get_param('include_choices_for', [])
        if choices_for:
            for field_name, field in self._get_readable_fields(
                    obj, field_names=choices_for):
                if base_hasattr(field, "value_type"):
                    vocab_name = field.value_type.vocabularyName
                else:
                    vocab_name = field.vocabularyName
                if vocab_name is None:
                    logger.warning(NO_VOCAB_ERROR % field_name)
                    continue
                vocab = get_vocab(obj, vocab_name)
                data_name = SUFFIXED_CHOICES_PATTERN % field_name
                result[data_name] = []
                for term in vocab._terms:
                    serializer = getMultiAdapter(
                        (term, self.request), interface=ISerializeToJson)
                    result[data_name].append(serializer())
        return result

    def _include_fields(self, obj):
        """ """
        result = {}
        # Compute fields if fullobjects or metadata_fields
        if self.fullobjects or self.metadata_fields:
            for field_name, field in self._get_readable_fields(
                    obj, field_names=self.metadata_fields or []):
                # serialize the field
                serializer = queryMultiAdapter(
                    (field, obj, self.request), IFieldSerializer
                )
                result[json_compatible(field_name)] = serializer()

        return result


@implementer(ISerializeToJson)
@adapter(IDexterityContainer, IPMRestapiLayer)
class BaseDXSerializeFolderToJson(BaseDXSerializeToJson, IMIODXSerializeFolderToJson):
    """ """



# -*- coding: utf-8 -*-

from plonemeeting.restapi.tests.base import BaseTestCase
from Products.PloneMeeting.tests.PloneMeetingTestCase import DEFAULT_USER_PASSWORD
from Products.PloneMeeting.utils import get_annexes

import transaction


class testServiceAnnexes(BaseTestCase):
    """ """

    def tearDown(self):
        self.api_session.close()

    def test_restapi_annexes_endpoint(self):
        """@annexes"""
        endpoint_url = "{0}/@annexes".format(self.portal_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        # nothing found for now
        self.assertEqual(response.json(), [])

        # create item with annexes
        self.changeUser("pmManager")
        item = self.create("MeetingItem")
        transaction.commit()

        # on MeetingItem without annexes
        item_url = item.absolute_url()
        endpoint_url = "{0}/@annexes".format(item_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.json(), [])
        # add annexes
        self.addAnnex(item)
        self.addAnnex(item)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(len(response.json()), 2)

    def test_restapi_annex_type_only_for_meeting_managers(self):
        """Make sure when a content_category is restricted to MeetingManagers,
           the same information is available in @annexes endpoint."""
        # make 'overhead-analysis' only_for_meeting_managers
        cfg = self.meetingConfig
        financial_analysis = cfg.annexes_types.item_annexes.get("financial-analysis")
        financial_analysis.only_for_meeting_managers = True
        # create item with annexes
        self.changeUser("pmManager")
        item = self.create("MeetingItem")
        item_url = item.absolute_url()
        endpoint_url = "{0}/@annexes?fullobjects".format(item_url)
        annex = self.addAnnex(item)
        transaction.commit()
        self.assertEqual(
            annex.content_category,
            "plonemeeting-assembly-annexes_types_-_item_annexes_-_financial-analysis",
        )
        response_pmManager = self.api_session.get(endpoint_url).json()[0]
        self.changeUser("pmCreator1")
        self.api_session.auth = ("pmCreator1", DEFAULT_USER_PASSWORD)
        # this will use the MissingTerms adapter used by annex.content_category
        response_pmCreator1 = self.api_session.get(endpoint_url).json()[0]
        self.assertEqual(
            response_pmManager["content_category"],
            response_pmCreator1["content_category"],
        )
        # in case value is really not found, we get correct format token/title
        # but with value as title
        annex.content_category = "some-removed-annex-type"
        transaction.commit()
        response_pmCreator1 = self.api_session.get(endpoint_url).json()[0]
        self.assertEqual(
            response_pmCreator1["content_category"],
            {u'title': u'Missing: some-removed-annex-type',
             u'token': u'some-removed-annex-type'}
        )

    def test_restapi_annexes_endpoint_filters(self):
        """@annexes, it is possible to filter on existing boolean values
           to_print, publishable, ..."""
        cfg = self.meetingConfig
        config = cfg.annexes_types.item_annexes
        config.publishable_activated = True
        # create item with annexes
        self.changeUser("pmManager")
        item = self.create("MeetingItem")
        annex = self.addAnnex(item, publishable=True)
        self.addAnnex(item, publishable=False)
        transaction.commit()
        item_url = item.absolute_url()
        endpoint_url = "{0}/@annexes?publishable=true".format(item_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual(len(get_annexes(item)), 2)
        annex_infos = response.json()[0]
        self.assertEqual(annex_infos['UID'], annex.UID())

        # check that it works with one single non publishable annex
        self.deleteAsManager(annex.UID())
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(len(response.json()), 0)

    def test_restapi_annexes_additional_values(self):
        """@annexes, additional_values will actually
        return categorized element infos."""
        self.changeUser("pmManager")
        item = self.create("MeetingItem")
        self.addAnnex(item)
        transaction.commit()
        item_url = item.absolute_url()
        # category_id
        endpoint_url = "{0}/@annexes?additional_values=category_id".format(item_url)
        response = self.api_session.get(endpoint_url)
        resp_json = response.json()
        self.assertFalse("category_title" in resp_json[0])
        self.assertEqual(resp_json[0]["category_id"], u'financial-analysis')
        # *
        endpoint_url = "{0}/@annexes?additional_values=*".format(
            item_url)
        response = self.api_session.get(endpoint_url)
        resp_json = response.json()
        self.assertTrue("category_id" in resp_json[0])
        self.assertTrue("category_title" in resp_json[0])
        self.assertTrue("category_uid" in resp_json[0])
        # there are some ignored values
        self.assertFalse("allowedRolesAndUsers" in resp_json[0])
        self.assertFalse("visible_for_groups" in resp_json[0])


def test_suite():
    from unittest import TestSuite, makeSuite

    suite = TestSuite()
    # add a prefix to avoid every PM tests to be executed
    suite.addTest(makeSuite(testServiceAnnexes, prefix="test_restapi_"))
    return suite



# -*- coding: utf-8 -*-
"""Installer for the plonemeeting.restapi package."""

from setuptools import find_packages
from setuptools import setup


long_description = "\n\n".join(
    [open("README.rst").read(), open("CHANGES.rst").read(), ])


setup(
    name="plonemeeting.restapi",
    version="2.1",
    description="Extended rest api service for Products.PloneMeeting usecases",
    long_description=long_description,
    # Get more from https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 4.3",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    ],
    keywords="Python Plone",
    author="Gauthier Bastien",
    author_email="gauthier@imio.be",
    url="https://pypi.python.org/pypi/plonemeeting.restapi",
    license="GPL version 2",
    packages=find_packages("src", exclude=["ez_setup"]),
    namespace_packages=["plonemeeting"],
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        # -*- Extra requirements: -*-
        "setuptools",
        "Products.PloneMeeting",
        "imio.restapi>=1.0a12",
        "plone.restapi>=7.8.0",
    ],
    extras_require={
        "test": ["plone.restapi[test]",
                 "Products.PloneMeeting[test]",
                 ], },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    [console_scripts]
    update_locale = plonemeeting.restapi.locales.update:update_locale
    """,
)



# -*- coding: utf-8 -*-

from plone.restapi.deserializer import boolean_value
from plone.restapi.interfaces import IExpandableElement
from plone.restapi.services import Service
from plonemeeting.restapi.config import ANNEXES_FILTER_VALUES
from plonemeeting.restapi.serializer.base import serialize_annexes
from Products.PloneMeeting.interfaces import IMeetingContent
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(IExpandableElement)
@adapter(IMeetingContent, Interface)
class Annexes(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        result = {"annexes": {"@id": "{}/@annexes".format(self.context.absolute_url())}}
        if not expand:
            return result

        # extend batch? DEFAULT_BATCH_SIZE = 25
        # self.request.form['b_size'] = 50

        # get filters from request.form
        # it is possible to filter annexes on every boolean attributes :
        # to_print, confidential, publishable, to_sign/signed
        filters = {k: boolean_value(v) for k, v in self.request.form.items()
                   if k in ANNEXES_FILTER_VALUES}
        result = serialize_annexes(self.context, filters)
        return result


class AnnexesGet(Service):
    def reply(self):
        annexes = Annexes(self.context, self.request)
        return annexes(expand=True)



# -*- coding: utf-8 -*-

from collective.iconifiedcategory.utils import calculate_category_id
from imio.helpers.content import get_vocab
from imio.helpers.security import fplog
from imio.restapi.services.add import FolderPost
from imio.restapi.utils import get_return_fullobject_after_creation_default
from plone import api
from plonemeeting.restapi.config import CONFIG_ID_ERROR
from plonemeeting.restapi.config import CONFIG_ID_NOT_FOUND_ERROR
from plonemeeting.restapi.utils import check_in_name_of
from plonemeeting.restapi.utils import rest_uuid_to_object
from Products.PloneMeeting.utils import add_wf_history_action
from Products.PloneMeeting.utils import get_annexes_config
from Products.PloneMeeting.utils import org_id_to_uid
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import NotFound


ANNEX_CONTENT_CATEGORY_ERROR = 'Given content_category "%s" was not found or ' \
    'is not useable for the annex you try to add!'
ANNEX_DECISION_RELATED_NOT_ITEM_ERROR = 'The "decision_related" parameter is ' \
    'only relevant when annex added on an item!'
ANNEX_NO_CONTENT_CATEGORY_AVAILABLE_ERROR = 'No annex_type available to create annex, ' \
    'please check configuration!'
IGNORE_VALIDATION_FOR_REQUIRED_ERROR = \
    'You can not ignore validation for following required fields: "%s"!'
IGNORE_VALIDATION_FOR_VALUED_ERROR = \
    'You can not ignore validation of a field for which a value ' \
    'is provided, remove "%s" from data!'
IGNORE_VALIDATION_FOR_WARNING = 'Validation was ignored for following fields: %s.'
OPTIONAL_FIELD_ERROR = 'The optional field "%s" is not activated in this configuration! ' \
    'You can ignore optional fields errors by adding "ignore_not_used_data: true" to sent data.'
OPTIONAL_FIELDS_WARNING = 'The following optional fields are not activated in ' \
    'this configuration and were ignored: %s.'
ORG_FIELD_VALUE_ERROR = 'Error with value "%s" defined for field "%s"! Enter a valid organization id or UID.'
UNKNOWN_DATA = "Following field names were ignored: %s."


class BasePost(FolderPost):

    # to be overrided
    required_fields = []

    def prepare_data(self, data):
        data = super(BasePost, self).prepare_data(data)

        # config_id
        self.tool = api.portal.get_tool("portal_plonemeeting")
        config_id = self._prepare_data_config_id(data)
        self.cfg = self.tool.get(config_id, None)
        if not self.cfg:
            raise BadRequest(CONFIG_ID_NOT_FOUND_ERROR % config_id)
        # type
        self.type = self._prepare_data_type(data)
        # main checks
        if data["@type"].startswith("Meeting"):
            # check that given values are useable, will raise if not
            data = self._check_optional_fields(data)
            # check fields for which validation will be ignored
            data = self._check_ignore_validation_for(data)
            # turn ids to UIDs
            data = self._turn_ids_into_uids(data)
            # fix data
            data = self._fix_data(data)
        elif data["@type"].startswith("annex"):
            # reinject data from parent: config_id and in_name_of
            data["config_id"] = self.cfg.getId()
            if not data.get("in_name_of", None) and self.parent_data.get(
               "in_name_of", None):
                data["in_name_of"] = self.parent_data["in_name_of"]
            # when adding a single annex, self is an AnnexPost instance
            # when adding an item with __children__ annexes it is not
            # the case but we use the _turn_ids_into_uids from AnnexPost
            # to cleanup data
            annex_post = self
            if not isinstance(annex_post, AnnexPost):
                portal = api.portal.get()
                annex_post = queryMultiAdapter(
                    (portal, self.request),
                    name="POST_application_json_@annex")
                annex_post._container = self.context
                annex_post.context = self.context
            data = annex_post._turn_ids_into_uids(data)
        return data

    def _prepare_data_config_id(self, data):
        if "config_id" not in data and "config_id" not in self.parent_data:
            raise BadRequest(CONFIG_ID_ERROR)
        return data.get("config_id", self.parent_data.get("config_id"))

    def _get_container(self):
        """ """
        return self.tool.getPloneMeetingFolder(self.cfg.getId())

    def _process_reply(self):
        # change context, the view is called on portal, when need
        # the set context to place where element will be added
        # if we have something else, probably we are adding an annex into an item or meeting
        if self.context.portal_type == "Plone Site":
            self.context = self._get_container()
        serialized_obj = super(BasePost, self)._reply()
        self._check_unknown_data(serialized_obj)
        return serialized_obj

    def _reply(self):
        in_name_of, access_cfg_ids = check_in_name_of(self.data['config_id'], self.data)
        if in_name_of:
            # remove AUTHENTICATED_USER during adopt_user
            auth_user = self.request.get("AUTHENTICATED_USER")
            if auth_user:
                self.request["AUTHENTICATED_USER"] = None
            with api.env.adopt_user(username=in_name_of):
                del self.data["in_name_of"]
                res = self._process_reply()
            if auth_user:
                self.request["AUTHENTICATED_USER"] = auth_user
            return res
        else:
            return self._process_reply()

    @property
    def _optional_fields(self):
        return []

    @property
    def _active_fields(self):
        return []

    def _check_optional_fields(self, data):
        optional_fields = self._optional_fields
        active_fields = self._active_fields
        inactive_fields = []
        ignore_not_used_data = data.get("ignore_not_used_data", False)
        for field_name, field_value in data.items():
            # if the field is an optional field that is not used
            # and that has a value (contains data), we warn if
            # ignore_not_used_data or we raise
            if (
                field_name in optional_fields and
                field_name not in active_fields and
                field_value
            ):
                if ignore_not_used_data:
                    inactive_fields.append(field_name)
                    data.pop(field_name)
                else:
                    raise BadRequest(OPTIONAL_FIELD_ERROR % field_name)

        if inactive_fields:
            self.warnings.append(OPTIONAL_FIELDS_WARNING %
                                 u', '.join(inactive_fields))
        return data

    def _check_ignore_validation_for(self, data):
        """Parameter "ignore_validation_for" will let ignore validation
           of given fields for which no value was provided."""
        ignore_validation_for = data.get("ignore_validation_for", [])
        if ignore_validation_for:
            # raise if trying to ignore validation of a required field
            ignored_required_fields = set(ignore_validation_for).intersection(
                self.required_fields)
            if ignored_required_fields:
                raise BadRequest(IGNORE_VALIDATION_FOR_REQUIRED_ERROR % u", ".join(
                    self.required_fields))

            # raise if trying to bypass validation and a value is given
            ignored_valued_fields = [k for k, v in data.items()
                                     if k in ignore_validation_for and v]
            if ignored_valued_fields:
                raise BadRequest(IGNORE_VALIDATION_FOR_VALUED_ERROR % u", ".join(
                    ignored_valued_fields))

            self.warnings.append(IGNORE_VALIDATION_FOR_WARNING %
                                 u', '.join(ignore_validation_for))
            # remove these fields from data to avoid setting a wrong empty value
            for field_name in ignore_validation_for:
                data.pop(field_name, None)
        return data

    @property
    def _turn_ids_into_uids_fieldnames(self):
        return []

    def _turn_ids_into_uids(self, data):
        # this will also include organizations outside "My org"
        org_uids = get_vocab(
            self.cfg,
            "Products.PloneMeeting.vocabularies.everyorganizationsvocabulary").by_token

        def _get_org_uid(field_name, field_value):
            """Get org UID as given field_value may be an UID or an id."""

            if field_value in org_uids:
                return field_value
            else:
                org_uid = org_id_to_uid(field_value, raise_on_error=False)
                if org_uid is None:
                    raise BadRequest(ORG_FIELD_VALUE_ERROR % (field_value, field_name))
                return org_uid

        for field_name in self._turn_ids_into_uids_fieldnames:
            field_value = data.get(field_name)
            if field_value:
                if hasattr(field_value, "__iter__"):
                    data[field_name] = []
                    for v in field_value:
                        data[field_name].append(_get_org_uid(field_name, v))
                else:
                    data[field_name] = _get_org_uid(field_name, field_value)
        return data

    def _fix_data(self, data):
        """Fix some data format."""
        # externalIdentifier must be string
        if "externalIdentifier" in data:
            externalIdentifier = data["externalIdentifier"]
            externalIdentifier = externalIdentifier and str(externalIdentifier) or ""
            data["externalIdentifier"] = externalIdentifier
        return data

    def clean_data(self, data):
        """Remove parameters that are not attributes."""
        cleaned_data = super(BasePost, self).clean_data(data)
        cleaned_data.pop("config_id", None)
        cleaned_data.pop("in_name_of", None)
        cleaned_data.pop("wf_transitions", None)
        cleaned_data.pop("ignore_not_used_data", None)
        cleaned_data.pop("decision_related", None)
        return cleaned_data

    def _after_reply_hook(self, serialized_obj):
        """ """
        obj = self.context.get(serialized_obj["id"])
        # add a record to the item workflow_history to specify that item was created thru SOAP WS
        action_name = "create_element_using_ws_rest"
        action_label = action_name + "_comments"
        # there may be several actions in the workflow_history, especially when
        # wf_transitions are used so we insert our event just after event 0
        add_wf_history_action(obj,
                              action_name=action_name,
                              action_label=action_label,
                              insert_index=0)
        # fingerpointing
        extras = "object={0}".format(repr(self.context.get(serialized_obj["id"])))
        fplog("create_by_ws_rest", extras=extras)

    def _check_unknown_data(self, serialized_obj):
        ignored = ["ignore_validation_for", "clean_html"]
        # if we have the full serialized obj, then we may check unknown data
        diff = []
        if get_return_fullobject_after_creation_default():
            diff = set(self.cleaned_data.keys()).difference(
                serialized_obj.keys() + ignored)
        if diff:
            self.warnings.append(UNKNOWN_DATA % ", ".join(diff))


class ItemPost(BasePost):

    required_fields = ["title", "proposingGroup"]

    @property
    def _turn_ids_into_uids_fieldnames(self):
        return [
            "proposingGroup",
            "groupsInCharge",
            "associatedGroups",
            "optionalAdvisers",
        ]

    @property
    def _optional_fields(self):
        return self.cfg.listUsedItemAttributes()

    @property
    def _active_fields(self):
        return self.cfg.getUsedItemAttributes()

    def _wf_transition_additional_warning(self, tr):
        warning_message = ""
        if tr == "present":
            warning_message = (
                " Make sure a meeting accepting items "
                "exists in configuration '{0}'!".format(self.cfg.getId())
            )
        return warning_message

    def _prepare_data_type(self, data):
        if not data.get("@type"):
            data["@type"] = self.cfg.getItemTypeName()
        return data.get("@type")


class MeetingPost(BasePost):

    required_fields = ["date"]

    @property
    def _optional_fields(self):
        return self.cfg.listUsedMeetingAttributes()

    @property
    def _active_fields(self):
        return self.cfg.getUsedMeetingAttributes()

    def _prepare_data_type(self, data):
        if not data.get("@type"):
            data["@type"] = self.cfg.getMeetingTypeName()
        return data.get("@type")


@implementer(IPublishTraverse)
class AnnexPost(BasePost):

    def __init__(self, context, request):
        super(AnnexPost, self).__init__(context, request)
        self.container_uid = None
        self._container = None

    def publishTraverse(self, request, name):
        if self.container_uid is None:
            self.container_uid = name
        else:
            raise NotFound(self, name, request)
        return self

    @property
    def container(self):
        if not self._container:
            self._container = rest_uuid_to_object(self.container_uid)
            self.context = self._container
        return self.context

    def _get_container(self):
        """ """
        return self.context

    def _prepare_data_config_id(self, data):
        tool = api.portal.get_tool("portal_plonemeeting")
        cfg = tool.getMeetingConfig(self.container)
        return cfg.getId()

    def _prepare_data_type(self, data):
        decision_related = data.get("decision_related", False)
        if decision_related and not self._container.__class__.__name__ == "MeetingItem":
            raise BadRequest(ANNEX_DECISION_RELATED_NOT_ITEM_ERROR)
        data["@type"] = decision_related and "annexDecision" or "annex"
        return data.get("@type")

    def _turn_ids_into_uids(self, data):
        # turn annex_type id into content_category calculated id
        content_category = data.get("content_category", None)
        annex_type = None
        # get selectable categories
        if data["@type"] == "annexDecision":
            self.request.set('force_use_item_decision_annexes_group', True)
        vocab = get_vocab(self.context, 'collective.iconifiedcategory.categories')
        # when content_category provided, it must exist
        if content_category is not None:
            annex_group = get_annexes_config(self.context, data["@type"], annex_group=True)
            # get given content_category or the first one
            annex_type = annex_group.get(content_category)
            if not annex_type:
                raise BadRequest(ANNEX_CONTENT_CATEGORY_ERROR % content_category)
            annex_type_value = calculate_category_id(annex_type)
            if annex_type_value not in vocab:
                raise BadRequest(ANNEX_CONTENT_CATEGORY_ERROR % content_category)
        else:
            # use the default annex_type, the first selectable
            if not vocab._terms:
                raise BadRequest(ANNEX_NO_CONTENT_CATEGORY_AVAILABLE_ERROR)
            annex_type_value = vocab._terms[0].token
        data["content_category"] = annex_type_value
        # cleanup
        if data["@type"] == "annexDecision":
            self.request.set('force_use_item_decision_annexes_group', False)

        return data



# -*- coding: utf-8 -*-

from imio.restapi.services.search import SearchGet
from plone import api
from plone.restapi.deserializer import boolean_value
from plonemeeting.restapi.config import CONFIG_ID_ERROR
from plonemeeting.restapi.config import CONFIG_ID_NOT_FOUND_ERROR
from plonemeeting.restapi.config import INDEX_CORRESPONDENCES
from plonemeeting.restapi.utils import check_in_name_of
from zExceptions import BadRequest


class BaseSearchGet(SearchGet):
    """ """

    config_id_optional = True

    def __init__(self, context, request):
        """ """
        super(BaseSearchGet, self).__init__(context, request)
        self.tool = api.portal.get_tool("portal_plonemeeting")
        self.type = self._type
        self.config_id = self._config_id

    @property
    def _config_id(self):
        if not self.config_id_optional and \
           ("config_id" not in self.request.form and
                "getConfigId" not in self.request.form):
            raise BadRequest(CONFIG_ID_ERROR)
        return self.request.form.get("config_id") or \
            self.request.form.get("getConfigId")

    @property
    def _type(self):
        return self.request.form.get("type")

    def _set_query_before_hook(self):
        """ """
        query = {}
        form = self.request.form

        # fix getConfigId to be unable to search outside self.access_cfg_ids
        if self.access_cfg_ids:
            query['getConfigId'] = self.access_cfg_ids

        # turn convenience indexes names into real index names
        for real_index_name, easy_index_name in INDEX_CORRESPONDENCES.items():
            if easy_index_name in form:
                query[real_index_name] = form[easy_index_name]

        return query

    def reply(self):
        """Override to handle in_name_of."""
        self.in_name_of, self.access_cfg_ids = check_in_name_of(self.config_id, self.request.form)
        if self.in_name_of:
            # remove AUTHENTICATED_USER during adopt_user
            auth_user = self.request.get("AUTHENTICATED_USER")
            if auth_user:
                self.request["AUTHENTICATED_USER"] = None
            with api.env.adopt_user(username=self.in_name_of):
                res = self._process_reply()
            if auth_user:
                self.request["AUTHENTICATED_USER"] = auth_user
            return res
        else:
            return self._process_reply()

    def _clean_query(self, query):
        """Remove or handle parameters that are not indexes names to avoid warnings
           like : WARNING plone.restapi.search.query No such index: 'extra_include'"""
        query.pop("extra_include", None)
        query.pop("in_name_of", None)
        for easy_index_name in INDEX_CORRESPONDENCES.values():
            query.pop(easy_index_name, None)


class PMSearchGet(BaseSearchGet):
    """Ease search of items and meetings related to a MeetingConfig.
       If config_id in REQUEST, we will configure some shortcuts,
       if not, the endpoint will behave as default."""

    def __init__(self, context, request):
        super(PMSearchGet, self).__init__(context, request)
        # if config_id is given, config_id must exist
        if self.config_id:
            self.cfg = self.tool.get(self.config_id, None)
            if not self.cfg:
                raise BadRequest(CONFIG_ID_NOT_FOUND_ERROR % self.config_id)

    @property
    def _config_id(self):
        config_id = self.request.form.get("config_id", None)
        in_name_of = self.request.form.get("in_name_of", None)
        # config_id or in_name_of is required when self.type is "item" or "meeting"
        if config_id is None and in_name_of is None and self.type in ["item", "meeting"]:
            raise BadRequest(CONFIG_ID_ERROR)
        return config_id

    @property
    def _type(self):
        # when using "config_id", default searched type is "item"
        if "type" not in self.request.form and \
           "config_id" in self.request.form:
            self.request.form["type"] = "item"
        return self.request.form.get("type")

    def _set_query_before_hook(self):
        query = super(PMSearchGet, self)._set_query_before_hook()
        query.update(self._set_query_meetings_accepting_items())
        return query

    def _set_query_meetings_accepting_items(self):
        """ """
        query = {}
        form = self.request.form
        if boolean_value(form.get("meetings_accepting_items", False)):
            # when using meetings_accepting_items, a config_id is required
            if not self.config_id:
                raise BadRequest(CONFIG_ID_ERROR)
            query.update(self.cfg._getMeetingsAcceptingItemsQuery())
            # we can call endpoint with just meetings_accepting_items=1
            self.type = "meeting"
        return query

    def _set_query_additional_params(self):
        """ """
        query = super(PMSearchGet, self)._set_query_additional_params()
        form = self.request.form
        if hasattr(self, "cfg") and self.cfg:
            mConfigs = [self.cfg]
        elif self.access_cfg_ids:
            mConfigs = [self.tool.get(e) for e in self.access_cfg_ids]

        # extend batch? DEFAULT_BATCH_SIZE = 25
        # self.request.form['b_size'] = 50

        # setup portal_type based on received 'type' parameter
        if self.type == "item":
            query["portal_type"] = [cfg.getItemTypeName() for cfg in mConfigs]
            query["sort_on"] = form.get("sort_on", "sortable_title")
        elif self.type == "meeting":
            query["portal_type"] = [cfg.getMeetingTypeName() for cfg in mConfigs]
            query["sort_on"] = form.get("sort_on", "sortable_title")
            query["sort_order"] = form.get("sort_order", "reverse")
            query["sort_on"] = form.get("sort_on", "sortable_title")
        elif self.type is not None:
            query["portal_type"] = self.type
            query["sort_on"] = form.get("sort_on", "sortable_title")
        return query

    def _clean_query(self, query):
        """See docstring in BaseSearchGet."""
        super(PMSearchGet, self)._clean_query(query)

        query.pop("additional_values", None)
        query.pop("base_search_uid", None)
        query.pop("meetings_accepting_items", None)

        # remove every values starting with include_ or extra_include_
        for k in query.keys():
            if k.startswith(('include_', 'extra_include_')):
                query.pop(k, None)
        # remove either "linkedMeetingUID" or "meeting_uid"
        catalog = self.context.portal_catalog
        if "linkedMeetingUID" in catalog.Indexes:
            query.pop("meeting_uid", None)
        else:
            query.pop("linkedMeetingUID", None)



# -*- coding: utf-8 -*-

from plone.restapi.services.users.get import UsersGet
from plonemeeting.restapi.utils import get_poweraccess_configs
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class PMUsersGet(UsersGet):
    def __init__(self, context, request):
        super(UsersGet, self).__init__(context, request)
        self.params = []
        self.query = self.request.form.copy()

    def has_permission_to_query(self):
        return get_poweraccess_configs()

    def has_permission_to_enumerate(self):
        return get_poweraccess_configs()

    def has_permission_to_access_user_info(self):
        return get_poweraccess_configs()



# -*- coding: utf-8 -*-

from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.serializer.atfields import DefaultFieldSerializer
from plone.restapi.serializer.atfields import TextFieldSerializer
from plone.restapi.serializer.converters import json_compatible
from plonemeeting.restapi.interfaces import IPMRestapiLayer
from plonemeeting.restapi.utils import handle_html
from Products.Archetypes.interfaces import IBaseObject
from Products.Archetypes.interfaces.field import ILinesField
from Products.Archetypes.interfaces.field import IStringField
from Products.Archetypes.interfaces.field import ITextField
from Products.CMFCore.utils import getToolByName
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


@adapter(IStringField, IBaseObject, Interface)
@implementer(IFieldSerializer)
class StringFieldSerializer(DefaultFieldSerializer):
    """Manage vocabulary for StringField."""
    def __call__(self):
        value = super(StringFieldSerializer, self).__call__()
        if self.field.vocabulary or self.field.vocabulary_factory:
            vocab = self.field.Vocabulary(self.context)
            value = {"token": value, "title": vocab.getValue(value)}
        return json_compatible(value)


@adapter(ILinesField, IBaseObject, Interface)
@implementer(IFieldSerializer)
class LinesFieldSerializer(DefaultFieldSerializer):
    """Manage vocabulary for LinesField."""
    def __call__(self):
        values = super(LinesFieldSerializer, self).__call__()
        if self.field.vocabulary or self.field.vocabulary_factory:
            result = []
            vocab = self.field.Vocabulary(self.context)
            for value in values:
                result.append({"token": value, "title": vocab.getValue(value)})
            values = result
        return json_compatible(values)


@adapter(ITextField, IBaseObject, IPMRestapiLayer)
@implementer(IFieldSerializer)
class PMTextFieldSerializer(TextFieldSerializer):
    """Manage images to data:image base64 values for text/html data."""
    def __call__(self):
        mimetypes_registry = getToolByName(self.context, "mimetypes_registry")
        data = super(TextFieldSerializer, self).__call__()
        content_type = json_compatible(mimetypes_registry(data)[2].normalized())
        if content_type == u'text/html':
            data = handle_html(self.context, data)
        return {
            "content-type": json_compatible(content_type),
            "data": data,
        }



# -*- coding: utf-8 -*-

from datetime import datetime
from imio.helpers.content import richtextval
from plone.app.textfield.value import RichTextValue
from plonemeeting.restapi.config import CONFIG_ID_ERROR
from plonemeeting.restapi.config import CONFIG_ID_NOT_FOUND_ERROR
from plonemeeting.restapi.tests.base import BaseTestCase
from plonemeeting.restapi.utils import IN_NAME_OF_UNAUTHORIZED
from plonemeeting.restapi.utils import IN_NAME_OF_USER_NOT_FOUND
from Products.PloneMeeting.MeetingItem import MeetingItem
from Products.PloneMeeting.tests.PloneMeetingTestCase import DEFAULT_USER_PASSWORD
from Products.PloneMeeting.tests.PloneMeetingTestCase import IMG_BASE64_DATA

import transaction


class testServiceSearch(BaseTestCase):
    """@search without 'type' is the same as @search?type=item"""

    def tearDown(self):
        self.api_session.close()

    def test_restapi_search_config_id_not_found(self):
        """Wrong config_id."""
        endpoint_url = "{0}/@search?config_id=unknown".format(self.portal_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {u"message": CONFIG_ID_NOT_FOUND_ERROR % "unknown", u"type": u"BadRequest"},
        )

    def test_restapi_search_config_id_error(self):
        """Parameter config_id is required when type is "item" or "meeting"."""
        # item
        endpoint_url = "{0}/@search?type=item".format(self.portal_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {u"message": CONFIG_ID_ERROR, u"type": u"BadRequest"},
        )
        # meeting
        endpoint_url = "{0}/@search?type=meeting".format(self.portal_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {u"message": CONFIG_ID_ERROR, u"type": u"BadRequest"},
        )
        # works for another type
        endpoint_url = "{0}/@search?type=Folder".format(self.portal_url)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json[u"items"][0][u"@type"], u"Folder")
        self.assertEqual(resp_json[u"items"][-1][u"@type"], u"Folder")

    def test_restapi_search_items_endpoint(self):
        """ """
        endpoint_url = "{0}/@search?config_id={1}&fullobjects=True".format(
            self.portal_url, self.meetingConfig.getId()
        )
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        # nothing found for now
        self.assertEqual(response.json()[u"items_total"], 0)

        # create 2 items
        self.changeUser("pmManager")
        item1 = self.create("MeetingItem")
        self.assertEqual(item1.query_state(), "itemcreated")
        item2 = self.create("MeetingItem")
        self.validateItem(item2)
        self.assertEqual(item2.query_state(), "validated")
        transaction.commit()

        # found
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 2)
        # may still use additional search parameters
        endpoint_url += "&review_state=validated"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json[u"items_total"], 1)
        self.assertEqual(resp_json[u"items"][0][u"review_state"], u"validated")

        # includes every data as well as extra formatted values
        self.assertTrue("motivation" in resp_json["items"][0])
        self.assertTrue("decision" in resp_json["items"][0])
        self.assertTrue("toDiscuss" in resp_json["items"][0])
        self.assertTrue("formatted_itemAssembly" in resp_json["items"][0])
        self.assertTrue("formatted_itemNumber" in resp_json["items"][0])
        self.assertTrue("all_copyGroups" in resp_json["items"][0])
        self.assertTrue("all_groupsInCharge" in resp_json["items"][0])
        transaction.abort()

    def test_restapi_search_items_in_meeting(self):
        """@search using the linkedMeetingUID attribute"""
        self.changeUser("pmManager")
        meeting = self._createMeetingWithItems()
        endpoint_url = (
            "{0}/@search?config_id={1}&linkedMeetingUID={2}"
            "&sort_on=getItemNumber".format(
                self.portal_url, self.meetingConfig.getId(), meeting.UID()
            )
        )
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        # items are returned sorted
        self.assertEqual(
            [elt["UID"] for elt in response.json()[u"items"]],
            [obj.UID() for obj in meeting.get_items(ordered=True)],
        )
        transaction.abort()

    def test_restapi_search_items_extra_include(self):
        """@search may receive an extra_include parameter"""
        transaction.begin()
        self.changeUser("pmManager")
        self._enableField(("category", "groupsInCharge"))
        self.getMeetingFolder()
        meeting = self.create("Meeting", date=datetime(2020, 6, 8, 8, 0))
        item = self.create("MeetingItem",
                           classifier="classifier1",
                           groupsInCharge=(self.developers_uid, ),
                           associatedGroups=(self.vendors_uid, ))
        item.setMotivation("<p>Motivation</p>")
        item.setDecision(self.decisionText)
        self.presentItem(item)
        endpoint_url = (
            "{0}/@search?config_id={1}&linkedMeetingUID={2}"
            "&sort_on=getItemNumber".format(
                self.portal_url, self.meetingConfig.getId(), meeting.UID()
            )
        )
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        # by default no extra include
        self.assertFalse("extra_include_proposing_group" in response.json()["items"][0])
        # does work even when fullobjects is not used
        endpoint_url = endpoint_url + "&extra_include=proposing_group"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertTrue("extra_include_proposing_group" in response.json()["items"][0])
        # now with fullobjects
        endpoint_url = endpoint_url + "&fullobjects"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertTrue("extra_include_proposing_group" in json["items"][0])
        self.assertFalse("extra_include_category" in json["items"][0])
        # extra_include proposing_group and category
        endpoint_url = endpoint_url + "&extra_include=category"
        transaction.begin()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertEqual(
            json["items"][0]["extra_include_proposing_group"]["id"], u"developers"
        )
        self.assertEqual(
            json["items"][0]["extra_include_category"]["id"], u"development"
        )
        # extra_include deliberation
        response = self.api_session.get(endpoint_url + "&extra_include=deliberation")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            response.json()["items"][0]["extra_include_deliberation"],
            {
                u"deliberation": u"<p>Motivation</p><p>Some decision.</p>",
            },
        )
        # extra_include several deliberation variants, but not "deliberation"
        endpoint_url = endpoint_url + "&extra_include=public_deliberation"
        endpoint_url = endpoint_url + "&extra_include=public_deliberation_decided"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            response.json()["items"][0]["extra_include_deliberation"],
            {
                u"public_deliberation": u"<p>Motivation</p><p>Some decision.</p>",
                u"public_deliberation_decided": u"<p>Motivation</p><p>Some decision.</p>",
            },
        )
        # extra_include meeting, need to pass also
        # extra_include_meeting_additional_values=*
        # to get additional_values like "formatted_date"
        endpoint_url = endpoint_url + \
            "&extra_include=meeting&extra_include_meeting_additional_values=*"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(
            resp_json["items"][0]["extra_include_meeting"]["UID"],
            meeting.UID()
        )
        self.assertEqual(
            resp_json["items"][0]["extra_include_meeting"]["formatted_date"],
            u'08/06/2020 (08:00)'
        )
        self.assertEqual(
            resp_json["items"][0]["extra_include_meeting"]["formatted_date_short"],
            u'08/06/2020'
        )
        self.assertEqual(
            resp_json["items"][0]["extra_include_meeting"]["formatted_date_long"],
            u'08 june 2020 (08:00)'
        )
        # extra_include everything
        endpoint_url = endpoint_url + "&extra_include=classifier"
        endpoint_url = endpoint_url + "&extra_include=groups_in_charge"
        endpoint_url = endpoint_url + "&extra_include=associated_groups"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        # by default, extra_include are summary
        self.assertFalse("@components" in resp_json["items"][0]["extra_include_category"])
        self.assertFalse("@components" in resp_json["items"][0]["extra_include_classifier"])
        self.assertFalse("@components" in resp_json["items"][0]["extra_include_meeting"])
        self.assertFalse("@components" in resp_json["items"][0]["extra_include_proposing_group"])
        self.assertFalse("@components" in resp_json["items"][0]["extra_include_groups_in_charge"])
        self.assertFalse("@components" in resp_json["items"][0]["extra_include_associated_groups"])
        # extra_include_fullobjects
        endpoint_url = endpoint_url + "&extra_include_category_fullobjects"
        endpoint_url = endpoint_url + "&extra_include_classifier_fullobjects"
        endpoint_url = endpoint_url + "&extra_include_meeting_fullobjects"
        endpoint_url = endpoint_url + "&extra_include_proposing_group_fullobjects"
        endpoint_url = endpoint_url + "&extra_include_groups_in_charge_fullobjects"
        endpoint_url = endpoint_url + "&extra_include_associated_groups_fullobjects"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_category"])
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_classifier"])
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_proposing_group"])
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_groups_in_charge"][0])
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_associated_groups"][0])
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_meeting"])
        self.assertEqual(resp_json["items"][0]["extra_include_classifier"]["id"],
                         "classifier1")
        self.assertEqual(resp_json["items"][0]["extra_include_groups_in_charge"][0]["id"],
                         "developers")
        self.assertEqual(resp_json["items"][0]["extra_include_associated_groups"][0]["id"],
                         "vendors")

        # for meeting moreover by default include_items=False
        self.assertFalse("items" in resp_json["items"][0]["extra_include_meeting"])
        endpoint_url = endpoint_url + "&extra_include_meeting_include_items=true"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertTrue("@components" in resp_json["items"][0]["extra_include_meeting"])
        self.assertTrue("items" in resp_json["items"][0]["extra_include_meeting"])
        transaction.abort()

    def test_restapi_search_items_extra_include_deliberation_images(self):
        """When asking for "deliberation" values, images are data base64 values"""
        transaction.begin()
        self.changeUser("pmManager")
        self._enableField('category')
        self.getMeetingFolder()
        item = self.create("MeetingItem")
        item.setMotivation("<p>Motivation</p>")
        img = self._add_image(item)
        pattern = '<p>Text with image <img src="{0}"/> and more text.</p>'
        text = pattern.format(img.absolute_url())
        item.setDecision(text)
        endpoint_url = "{0}/@search?config_id={1}&type=item&fullobjects" \
            "&extra_include=deliberation_decision".format(
                self.portal_url, self.meetingConfig.getId())
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(
            resp_json["items"][0]["extra_include_deliberation"]["deliberation_decision"],
            pattern.format(IMG_BASE64_DATA))

    def test_restapi_search_meetings_endpoint(self):
        """ """
        endpoint_url = "{0}/@search?config_id={1}&type=meeting".format(
            self.portal_url, self.meetingConfig.getId()
        )
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        # nothing found for now
        self.assertEqual(response.json()[u"items_total"], 0)

        # create 2 meetings
        self.changeUser("pmManager")
        meeting = self.create("Meeting", date=datetime(2018, 11, 18))
        self.assertEqual(meeting.query_state(), "created")
        meeting2 = self.create("Meeting", date=datetime(2019, 11, 18))
        self.closeMeeting(meeting2)
        self.assertEqual(meeting2.query_state(), "closed")
        transaction.commit()

        # found
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json[u"items_total"], 2)
        # meetings are sorted by date, from newest to oldest
        self.assertEqual(
            [m["title"] for m in resp_json[u"items"]],
            [u"18 november 2019", u"18 november 2018"],
        )
        # may still use additional search parameters
        endpoint_url += "&review_state=closed"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 1)
        self.assertEqual(response.json()[u"items"][0][u"review_state"], u"closed")
        transaction.abort()

    def test_restapi_search_meetings_fullobjects(self):
        """ """
        endpoint_url = "{0}/@search?config_id={1}&type=meeting&fullobjects=True".format(
            self.portal_url, self.meetingConfig.getId()
        )

        # create 2 meetings
        self.changeUser("pmManager")
        pattern = '<p>Text with image <img src="{0}"/> and more text.</p>'
        meeting = self.create("Meeting", date=datetime(2019, 11, 18))
        meeting2 = self.create("Meeting", date=datetime(2019, 11, 19))
        meeting2.assembly = RichTextValue(u'Mr Present, [[Mr Absent]], Mr Present2')
        img = self._add_image(meeting2)
        text = pattern.format(img.absolute_url())
        meeting2.observations = richtextval(text)
        self.assertEqual(meeting.query_state(), "created")
        self.closeMeeting(meeting2)
        self.assertEqual(meeting2.query_state(), "closed")
        transaction.commit()

        # found
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 2)
        # may still use additional search parameters
        endpoint_url += "&review_state=closed"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json[u"items_total"], 1)
        self.assertEqual(resp_json[u"items"][0][u"review_state"], u"closed")

        # includes every data as well as extra formatted values
        self.assertTrue("date" in resp_json["items"][0])
        # AT/DX
        self.assertTrue("startDate" in resp_json["items"][0] or
                        "start_date" in resp_json["items"][0])
        self.assertTrue("notes" in resp_json["items"][0])
        self.assertEqual(
            resp_json["items"][0]["formatted_assembly"],
            u'<p>Mr Present, <strike>Mr Absent</strike>, Mr Present2</p>')
        self.assertEqual(resp_json["items"][0]["observations"]["data"],
                         pattern.format(IMG_BASE64_DATA))
        transaction.abort()

    def test_restapi_search_meetings_accepting_items(self):
        """ """
        cfg = self.meetingConfig
        endpoint_url = "{0}/@search?config_id={1}&type=meeting".format(
            self.portal_url, cfg.getId()
        )

        # create 2 meetings
        self.changeUser("pmManager")
        meeting = self.create("Meeting", date=datetime(2020, 5, 10))
        self.assertEqual(meeting.query_state(), "created")
        meeting2 = self.create("Meeting", date=datetime(2020, 5, 17))
        self.closeMeeting(meeting2)
        self.assertEqual(meeting2.query_state(), "closed")
        transaction.commit()
        # only meeting is accepting items
        self.assertEqual(
            [meeting.UID()], [brain.UID for brain in cfg.getMeetingsAcceptingItems()]
        )

        # both found by default
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 2)
        endpoint_url += "&meetings_accepting_items=True"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertEqual(json[u"items_total"], 1)

        # type=meeting is optional
        endpoint_url = endpoint_url.replace("&type=meeting", "")
        response = self.api_session.get(endpoint_url)
        # excepted the @id that contains the query, the response is the same
        new_json = response.json()
        json.pop('@id')
        new_json.pop('@id')
        self.assertEqual(json, new_json)

        # we can ask for more details, like meeting date
        self.assertFalse("date" in json["items"][0])
        endpoint_url += "&metadata_fields=date"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertEqual(json[u"items_total"], 1)
        self.assertEqual(json["items"][0]["date"], u'2020-05-10T00:00:00')

    def test_restapi_search_in_name_of(self):
        """ """
        cfg = self.meetingConfig
        endpoint_url_pattern = "{0}/@search?config_id={1}&in_name_of=%s".format(
            self.portal_url, cfg.getId()
        )
        # must be (Meeting)Manager to use in_name_of
        self.api_session.auth = ("pmCreator1", DEFAULT_USER_PASSWORD)
        response = self.api_session.get(endpoint_url_pattern % "pmCreator2")
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.json(),
            {
                u"message": IN_NAME_OF_UNAUTHORIZED % ("pmCreator1", "pmCreator2"),
                u"type": u"Unauthorized",
            },
        )
        # user must exist
        self.api_session.auth = ("pmManager", DEFAULT_USER_PASSWORD)
        response = self.api_session.get(endpoint_url_pattern % "unknown")
        self.assertEqual(
            response.json(),
            {u"message": IN_NAME_OF_USER_NOT_FOUND % "unknown", u"type": u"BadRequest"},
        )
        # create 2 items, one for developers, one for vendors
        self.api_session.auth = ("pmManager", DEFAULT_USER_PASSWORD)
        self._addPrincipalToGroup("pmManager", self.vendors_creators)
        self.changeUser("pmManager")
        item1 = self.create("MeetingItem", proposingGroup=self.developers_uid)
        item2 = self.create("MeetingItem", proposingGroup=self.vendors_uid)
        transaction.commit()
        # can not pass an empty in_name_of
        response = self.api_session.get(endpoint_url_pattern % "")
        self.assertEqual(
            response.json(),
            {u"message": IN_NAME_OF_USER_NOT_FOUND % "", u"type": u"BadRequest"},
        )
        # both found by default
        response = self.api_session.get((endpoint_url_pattern % "").replace('&in_name_of=', ''))
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 2)
        # as pmCreator1, only item1 found
        response = self.api_session.get(endpoint_url_pattern % "pmCreator1")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 1)
        self.assertEqual(response.json()["items"][0][u"UID"], item1.UID())
        # as pmCreator2, only item2 found
        response = self.api_session.get(endpoint_url_pattern % "pmCreator2")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 1)
        self.assertEqual(response.json()["items"][0][u"UID"], item2.UID())

    def test_restapi_search_in_name_of_accross_configs(self):
        """Ensure that when in_name_of parameter is used but no config_id is given,
        the search is made accross all user configs"""
        self.api_session.auth = ("pmManager", DEFAULT_USER_PASSWORD)
        endpoint_url_pattern = "{0}/@search?in_name_of={1}&type=item".format(
            self.portal_url, "pmCreator2"
        )
        self._addPrincipalToGroup("pmManager", self.vendors_creators)
        self.changeUser("pmManager")
        item1 = self.create("MeetingItem", proposingGroup=self.vendors_uid)
        self.setMeetingConfig(self.cfg2_id)
        item2 = self.create("MeetingItem", proposingGroup=self.vendors_uid)
        transaction.commit()
        response = self.api_session.get(endpoint_url_pattern)
        self.assertEqual(response.status_code, 200, response.content)
        result = response.json()
        self.assertEqual(2, result["items_total"])
        self.assertEqual(
            sorted([item1.UID(), item2.UID()]),
            sorted([r["UID"] for r in result["items"]]),
        )
        self.assertEqual(
            sorted(["MeetingItemPma", "MeetingItemPga"]),
            sorted([r["@type"] for r in result["items"]]),
        )

    def test_restapi_search_base_search_uid(self):
        """A collection UID may be given as base search."""
        cfg = self.meetingConfig
        # use the searchdecideditems collection
        base_search_uid = self.meetingConfig.searches.searches_items.searchmyitems.UID()
        endpoint_url = "{0}/@search?config_id={1}&base_search_uid={2}".format(
            self.portal_url, cfg.getId(), base_search_uid
        )
        # nothing found for now
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(response.json()[u"items_total"], 0)
        # create one item for developers and one for vendors as pmManager
        # and one item as pmCreator1
        self.changeUser("pmCreator1")
        item_dev1 = self.create("MeetingItem", proposingGroup=self.developers_uid)
        item_dev1_uid = item_dev1.UID()
        self._addPrincipalToGroup("pmManager", self.vendors_creators)
        self.changeUser("pmManager")
        item_dev2 = self.create("MeetingItem", proposingGroup=self.developers_uid)
        item_dev2_uid = item_dev2.UID()
        item_ven1 = self.create("MeetingItem", proposingGroup=self.vendors_uid)
        item_ven1_uid = item_ven1.UID()
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        # only 2 items found
        self.assertEqual(json[u"items_total"], 2)
        self.assertEqual(json[u"items"][0]["UID"], item_dev2_uid)
        self.assertEqual(json[u"items"][1]["UID"], item_ven1_uid)
        # possible to complete the query with arbitray parameters
        # restrict only developers_uid
        endpoint_url += "&getProposingGroup={0}".format(self.developers_uid)
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        # only 1 item found
        self.assertEqual(json[u"items_total"], 1)
        self.assertEqual(json[u"items"][0]["UID"], item_dev2_uid)
        # override Creator, use pmCreator1
        endpoint_url += "&Creator=pmCreator1"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertEqual(json[u"items_total"], 1)
        self.assertEqual(json[u"items"][0]["UID"], item_dev1_uid)
        # get items from pmManager and pmCreator1
        endpoint_url += "&Creator=pmManager"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertEqual(json[u"items_total"], 2)
        self.assertEqual(json[u"items"][0]["UID"], item_dev1_uid)
        self.assertEqual(json[u"items"][1]["UID"], item_dev2_uid)

    def test_restapi_search_without_config_id(self):
        """@search parameter config_id will ease searching but
           when not provided, then default @search functionnality is available."""
        self.changeUser("pmManager")
        endpoint_url = (
            "{0}/@search?portal_type=organization".format(self.portal_url)
        )
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        # organizations are returned
        json = response.json()
        for result in json[u"items"]:
            self.assertEqual(result[u'@type'], 'organization')
        transaction.abort()

    def test_restapi_search_without_type(self):
        """The @search endpoint original behavior is still useable."""
        cfg = self.meetingConfig
        self.changeUser("pmManager")
        endpoint_url = (
            "{0}/@search?UID={1}".format(self.portal_url, cfg.UID())
        )
        transaction.commit()
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        json = response.json()
        self.assertEqual(json[u"items_total"], 1)
        self.assertEqual(json[u"items"][0][u'id'], cfg.getId())

    def test_restapi_search_fullobjects_and_includes(self):
        """Several includes may be passed as parameter when "fullobjects" is used.
           By default, "fullobjects" behavior is like in plone.restapi, it retuens every infos.
           But we may pass additional includes:
           - fullobjects: when given, every is returned, if not only base data is returnerd;
           - include_base_data: will include base data like id, uid, review_state, ...;
           - include_components: will include the "@components" section;
           - include_nextprev: will include next/previous infos;
           - include_parent: will serialize parent into "parent" key;
           - include_items: will include the contained elements;
           - include_target_url: will include the "targetUrl";
           - include_allow_discussion: will include infos about discussion;
           - additional_values: will include given additional values.
        """
        # create 1 item
        self.changeUser("pmManager")
        self.create("MeetingItem")
        transaction.commit()

        endpoint_url = "{0}/@search?config_id={1}&fullobjects=True".format(
            self.portal_url, self.meetingConfig.getId()
        )
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json[u"items_total"], 1)
        # by default everything is there, except "items" that is False by default
        self.assertTrue("@components" in resp_json["items"][0])
        self.assertTrue("id" in resp_json["items"][0])
        self.assertTrue("UID" in resp_json["items"][0])
        self.assertTrue("next_item" in resp_json["items"][0])
        self.assertTrue("previous_item" in resp_json["items"][0])
        # parent is only there when specifically asked
        self.assertFalse("parent" in resp_json["items"][0])
        self.assertTrue("allow_discussion" in resp_json["items"][0])
        self.assertTrue("layout" in resp_json["items"][0])
        self.assertTrue("formatted_itemNumber" in resp_json["items"][0])
        self.assertFalse("items" in resp_json["items"][0])
        # we may get what we want, only get "@components"
        endpoint_url = "{0}/@search?config_id={1}" \
            "&include_base_data=false&include_parent=true" \
            "&include_components=true".format(
                self.portal_url, self.meetingConfig.getId())
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json[u"items_total"], 1)
        self.assertTrue("@components" in resp_json["items"][0])
        self.assertFalse("id" in resp_json["items"][0])
        self.assertFalse("UID" in resp_json["items"][0])
        self.assertFalse("next_item" in resp_json["items"][0])
        self.assertFalse("previous_item" in resp_json["items"][0])
        # include_parent=true
        self.assertTrue("parent" in resp_json["items"][0])
        self.assertFalse("allow_discussion" in resp_json["items"][0])
        self.assertFalse("formatted_itemNumber" in resp_json["items"][0])
        self.assertFalse("items" in resp_json["items"][0])

    def test_restapi_search_extra_includes_parameters(self):
        """Every parameters may be passed to extra_includes:
           Example: extra_include=category, enable fullobjects and include_base_data=false:
           Parameters would be the following:
           - extra_include=category;
           - extra_include_category_fullobjects;
           - extra_include_category_include_base_data=false.
        """
        self._enableField('category')
        # create 2 items
        self.changeUser("pmManager")
        self.create("MeetingItem")
        transaction.commit()

        endpoint_url = "{0}/@search?config_id={1}&extra_include=category" \
            "&extra_include_category_include_components=true".format(
                self.portal_url, self.meetingConfig.getId())
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        # we get @components and base data
        self.assertEqual(sorted(resp_json["items"][0]["extra_include_category"].keys()),
                         [u'@components',
                          u'@extra_includes',
                          u'@id',
                          u'@type',
                          u'UID',
                          u'created',
                          u'enabled',
                          u'id',
                          u'modified',
                          u'review_state',
                          u'title'])

    def test_restapi_search_metadata_fields(self):
        """metadata_fields may be used:
           - without fullobjects, the metadata catalog is used;
           - with fullobjects, then it is used to select the fields we want.
        """
        self._enableField('category')
        # create 2 items
        self.changeUser("pmManager")
        self.create("MeetingItem")
        transaction.commit()

        # get item base data + getItemNumber and category enabled and category_id
        endpoint_url = "{0}/@search?config_id={1}" \
            "&metadata_fields=itemNumber" \
            "&extra_include=category" \
            "&extra_include_category_include_base_data=false" \
            "&extra_include_category_metadata_fields=enabled" \
            "&extra_include_category_metadata_fields=category_id".format(
                self.portal_url, self.meetingConfig.getId())
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        # item
        self.assertEqual(sorted(resp_json["items"][0].keys()),
                         [u'@extra_includes', u'@id', u'@type', u'UID',
                          u'created', u'extra_include_category', u'id',
                          u'itemNumber', u'modified', u'review_state', u'title'])
        self.assertEqual(resp_json["items"][0]["itemNumber"], 0)
        # category
        self.assertEqual(resp_json["items"][0]["extra_include_category"],
                         {u'@extra_includes': [],
                          u'category_id': u'development',
                          u'enabled': True})

    def test_restapi_search_not_found_brain(self):
        """Check that everything is behaving correctly when some brains
        are lost in portal_catalog (object does not exist anymore)."""
        self.changeUser("pmManager")
        item1 = self.create("MeetingItem")
        item1_uid = item1.UID()
        item2 = self.create("MeetingItem")
        # prevent unindex when deleting item1
        old__unindexObject = MeetingItem.unindexObject
        MeetingItem.unindexObject = lambda *args: None
        self.deleteAsManager(item1_uid)
        MeetingItem.unindexObject = old__unindexObject
        self.assertTrue(self.catalog(UID=item1_uid))
        self.assertFalse(item1 in item1.aq_parent.objectValues())
        self.assertTrue(item2 in item1.aq_parent.objectValues())
        transaction.commit()

        # when using brains, 2 results
        endpoint_url = "{0}/@search?config_id={1}&sort_on=getId".format(
            self.portal_url, self.meetingConfig.getId())
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(len(resp_json["items"]), 2)
        # with fullobjects, only 1 result
        endpoint_url += "&fullobjects"
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(len(resp_json["items"]), 1)

    def test_restapi_search_data_are_anonymized(self):
        """Data collected from PloneMeeting are anonymized."""
        self._enableField("observations")
        self._enableField("observations", related_to="Meeting")

        # create 2 items
        text = '<p>Text shown<span class="pm-anonymize"> text hidden</span> and ' \
            'some <span class="highlight-red">highlighted text</span>.</p>'
        anonymized_text = '<p>Text shown<span class="pm-anonymize"></span> and ' \
            'some <span class="highlight-red">highlighted text</span>.</p>'
        self.changeUser("pmManager")
        item = self.create("MeetingItem")
        item.setObservations(text)
        item.setDecision(text)
        # in 4.1.x item observations was only readable until validated...
        self.validateItem(item)
        meeting = self.create("Meeting", date=datetime(2020, 6, 8, 8, 0))
        meeting.observations = richtextval(text)
        transaction.commit()

        # query to get meeting and item description
        endpoint_url = "{0}/@search?extra_include=public_deliberation" \
            "&metadata_fields=observations" \
            "&UID={2}&UID={3}&sort_on=getId".format(
                self.portal_url, self.meetingConfig.getId(),
                item.UID(), meeting.UID())
        response = self.api_session.get(endpoint_url)
        self.assertEqual(response.status_code, 200, response.content)
        resp_json = response.json()
        self.assertEqual(resp_json["items_total"], 2)
        # item
        # RichTextField
        self.assertEqual(resp_json["items"][0]["observations"]["data"], anonymized_text)
        # RichText treated by printXhtml
        self.assertEqual(
            resp_json["items"][0]["extra_include_deliberation"]["public_deliberation"],
            anonymized_text)
        # meeting
        # RichTextField
        self.assertEqual(resp_json["items"][1]["observations"]["data"], anonymized_text)


def test_suite():
    from unittest import TestSuite, makeSuite

    suite = TestSuite()
    # add a prefix to avoid every PM tests to be executed
    suite.addTest(makeSuite(testServiceSearch, prefix="test_restapi_"))
    return suite



# -*- coding: utf-8 -*-

from plone import api
from plone.restapi.deserializer import json_body
from plone.restapi.deserializer.atcontent import DeserializeFromJson as ATDeserializeFromJson
from plone.restapi.interfaces import IDeserializeFromJson
from Products.PloneMeeting.interfaces import IATMeetingContent
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(IDeserializeFromJson)
@adapter(IATMeetingContent, Interface)
class DeserializeFromJson(ATDeserializeFromJson):
    """JSON deserializer for Archetypes content types
    """

    notifies_create = True

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def _need_update_local_roles(self, data):
        """Do we need to update_local_roles before setting data on element?"""
        return "internalNotes" in data

    def __call__(self, validate_all=False, data=None, create=False):

        if create:
            wfTool = api.portal.get_tool("portal_workflow")
            wf = wfTool.getWorkflowsFor(self.context)[0]
            wf.updateRoleMappingsFor(self.context)
            # some fields like "internalNotes" need local_roles to be set up
            # to be writable, so check if we need to update_local_roles
            # we do it only if required
            if data is None:
                data = json_body(self.request)
                if "proposingGroup" in data and self._need_update_local_roles(data):
                    # for now self.context does not have any field set, we need
                    # to set proposingGroup so update_local_roles works
                    self.context.setProposingGroup(data["proposingGroup"])
                    self.context.update_local_roles()

        return super(DeserializeFromJson, self).__call__(
            validate_all=validate_all, data=data, create=create)

    def validate(self):
        """Handle "ignore_validation_for" that will specifically
           let ignore validation errors for given fields."""
        errors = super(DeserializeFromJson, self).validate()
        data = json_body(self.request)
        ignore_validation_for = data.get("ignore_validation_for", [])
        for field_name in ignore_validation_for:
            errors.pop(field_name, None)
        return errors



# -*- coding: utf-8 -*-

from plone.app.testing import FunctionalTesting
from plone.restapi.testing import PLONE_RESTAPI_AT_FUNCTIONAL_TESTING
from Products.PloneMeeting.testing import PMLayer

import plonemeeting.restapi


class PMRestapiLayer(PMLayer):
    """ """


PM_REST_TESTING_PROFILE = PMRestapiLayer(
    zcml_filename="testing.zcml",
    zcml_package=plonemeeting.restapi,
    additional_z2_products=(
        "imio.dashboard",
        "Products.PloneMeeting",
        "Products.CMFPlacefulWorkflow",
        "Products.PasswordStrength",
    ),
    gs_profile_id="plonemeeting.restapi:testing",
    name="PM_REST_TESTING_PROFILE",
)


PM_REST_TEST_PROFILE_FUNCTIONAL = FunctionalTesting(
    bases=(PM_REST_TESTING_PROFILE, PLONE_RESTAPI_AT_FUNCTIONAL_TESTING),
    name="PM_REST_TEST_PROFILE_FUNCTIONAL",
)


PM_REST_TEST_ADD_PROFILE_FUNCTIONAL = FunctionalTesting(
    bases=(PM_REST_TESTING_PROFILE, PLONE_RESTAPI_AT_FUNCTIONAL_TESTING),
    name="PM_REST_TEST_ADD_PROFILE_FUNCTIONAL",
)

PM_REST_TEST_ADD_ANNEXES_PROFILE_FUNCTIONAL = FunctionalTesting(
    bases=(PM_REST_TESTING_PROFILE, PLONE_RESTAPI_AT_FUNCTIONAL_TESTING),
    name="PM_REST_TEST_ADD_ANNEXES_PROFILE_FUNCTIONAL",
)
